from PySide6.QtWidgets import QWidget, QTableWidgetItem, QLabel, QComboBox
from PySide6.QtCore import Qt
from nomad_camels.main_classes.loop_step import Loop_Step, Loop_Step_Config

from nomad_camels.gui.read_channels import Ui_read_channels_config

from nomad_camels.utility import variables_handling, fit_variable_renaming
from nomad_camels.ui_widgets.channels_check_table import Channels_Check_Table


class Read_Channels(Loop_Step):
    """
    This step represents the bluesky plan stub `trigger_and_read`. It may also
    be split into an additional step for triggering, then doing something else
    and then reading.

    Attributes
    ----------
    read_all : bool
        If True, the step will read all available channels.
    split_trigger : bool
        If True, an additional trigger channels step may be used. This read step
        will then not use `trigger_and_read`, but only read the channels.
    channel_list : list[str]
        The list of channels that should be read in this step.
    """

    def __init__(self, name="", parent_step=None, step_info=None, **kwargs):
        super().__init__(name, parent_step, step_info, **kwargs)
        self.step_type = "Read Channels"
        if step_info is None:
            step_info = {}
        self.read_all = step_info["read_all"] if "read_all" in step_info else False
        self.split_trigger = (
            step_info["split_trigger"] if "split_trigger" in step_info else False
        )
        if "channel_list" in step_info:
            self.channel_list = step_info["channel_list"]
        else:
            self.channel_list = []
        if "read_variables" in step_info:
            self.read_variables = step_info["read_variables"]
        else:
            self.read_variables = False
        self.update_used_devices()

    def update_used_devices(self):
        """All devices that should be read are added to the used_devices."""
        self.used_devices = []
        for channel in variables_handling.channels:
            if self.read_all or channel in self.channel_list:
                device = variables_handling.channels[channel].device
                if device not in self.used_devices:
                    self.used_devices.append(device)

    def get_channels_set(self):
        """Provides a set of `self.channel_list` to remove possible duplicates.
        Includes all available channels if `self.read_all`."""
        chan_list = []
        if self.read_all:
            for channel in variables_handling.channels:
                chan_list.append(channel)
        else:
            chan_list = self.channel_list
        return set(chan_list)

    def get_channels_string(self, tabs):
        """
        Gives a string of the channels that should be read. This may also be
        used by the Trigger_Channels step.

        Parameters
        ----------
        tabs : str
            A string including the tabs for intendation.
        """
        channel_string = f"{tabs}channels_{self.variable_name()} = ["
        if not self.read_all and not self.channel_list and not self.read_variables:
            raise Exception(f"Trying to read no channel in {self.full_name}!")
        if self.read_all:
            for channel in variables_handling.channels:
                channel_string += get_channel_string(channel)
        else:
            for channel in self.channel_list:
                channel_string += get_channel_string(channel)
        if self.read_variables:
            channel_string += f"{self.protocol.name}_variable_signal, "
        channel_string = channel_string[:-2] + "]\n"
        return channel_string

    def variable_name(self):
        """Returns the name of this step as a valid variable name, to specify
        the channels for this read."""
        return fit_variable_renaming.replace_name(self.name)

    def get_protocol_string(self, n_tabs=1):
        """In the protocol, at first a list `channels` is defined,
        including all the channels, that are selected to be read. Then
        `bps.trigger_and_read` (or `helper_functions.read_wo_trigger`) is called
        on these channels.
        The stream in which the data is written will be numbered if there are
        other read_channels that are reading different channels, since bluesky
        only allows reading the same channels inside one stream."""
        # checking compatibility with other readings
        chan_list = self.get_channels_set()
        channels_w_variables = set(list(chan_list) + [self.read_variables])
        if channels_w_variables in variables_handling.read_channel_sets:
            n = variables_handling.read_channel_sets.index(channels_w_variables)
        else:
            n = len(variables_handling.read_channel_names)
            variables_handling.read_channel_sets.append(channels_w_variables)
            if n > 0:
                variables_handling.read_channel_names.append(f'f"{{stream_name}}_{n}"')
            else:
                variables_handling.read_channel_names.append("stream_name")
        stream_name = variables_handling.read_channel_names[n]
        tabs = "\t" * n_tabs
        protocol_string = super().get_protocol_string(n_tabs)
        if self.split_trigger:
            protocol_string += f"{tabs}yield from helper_functions.read_wo_trigger(channels_{self.variable_name()}, grp_{self.variable_name()}, stream={stream_name})\n"
        else:
            protocol_string += self.get_channels_string(tabs)
            protocol_string += f"{tabs}yield from bps.trigger_and_read(channels_{self.variable_name()}, name={stream_name})\n"
        return protocol_string

    def get_protocol_short_string(self, n_tabs=0):
        """Includes the channel list in the string."""
        short_string = super().get_protocol_short_string(n_tabs)
        short_string = f"{short_string[:-1]} - {self.channel_list}\n"
        return short_string


def get_channel_string(channel):
    """
    Gives the string of a channel in the way it is written inside the
    protocol, i.e. "devs["<device_name>"].<component/channel_name>".

    Parameters
    ----------
    channel : str
        The channel that should be converted.
    """
    name = variables_handling.channels[channel].name
    if "." in name:
        dev, chan = name.split(".")
        return f'devs["{dev}"].{chan}, '
    else:
        return f'devs["{name}"], '


class Read_Channels_Config(Loop_Step_Config):
    """ """

    def __init__(self, loop_step: Read_Channels, parent=None):
        super().__init__(parent, loop_step)
        self.sub_widget = Read_Channels_Config_Sub(loop_step, self)
        self.layout().addWidget(self.sub_widget, 1, 0, 1, 5)

    def update_step_config(self):
        """ """
        super().update_step_config()
        self.sub_widget.update_step_config()


class Read_Channels_Config_Sub(Ui_read_channels_config, QWidget):
    """Config for the Read_Channels it provides a table of channels with
    a checkbox, whether to read them. Also there is a checkbox whether
    to simply read all available channels.

    Parameters
    ----------

    Returns
    -------

    """

    def __init__(self, loop_step: Read_Channels, parent=None):
        super().__init__(parent)
        self.setupUi(self)
        self.loop_step = loop_step
        self.checkBox_read_all.stateChanged.connect(self.read_type_changed)
        self.checkBox_split_trigger.stateChanged.connect(self.use_trigger)
        # self.comboBox_readType.addItems(['read all', 'read selected'])
        # self.comboBox_readType.currentTextChanged.connect(self.read_type_changed)
        self.load_data()
        labels = ["read", "channel"]
        info_dict = {"channel": self.loop_step.channel_list}
        self.read_table = Channels_Check_Table(
            self, labels, info_dict=info_dict, title="Read-Channels"
        )
        self.read_type_changed()
        self.layout().addWidget(self.read_table, 5, 0, 1, 3)
        # self.tableWidget_channels.setHorizontalHeaderLabels(['read',
        #                                                      'channel name',
        #                                                      'use set-value'])
        # self.build_channels_table()
        # self.lineEdit_search.textChanged.connect(self.build_channels_table)
        # self.checkBox_use_set.toggled.connect(self.checkbox_toggle)
        # self.checkBox_plot.toggled.connect(self.checkbox_toggle)
        # self.checkBox_save.toggled.connect(self.checkbox_toggle)
        # self.tableWidget_channels.clicked.connect(self.table_check_changed)

    def use_trigger(self):
        """ """
        self.loop_step.split_trigger = self.checkBox_split_trigger.isChecked()

    # def checkbox_toggle(self):
    #     """When a checkbox is (un-)checked, the new value is stored
    #     inside the loop_step."""
    #     self.loop_step.use_set_val = self.checkBox_use_set.isChecked()
    #     self.loop_step.save_data = self.checkBox_save.isChecked()
    #     self.loop_step.plot_data = self.checkBox_plot.isChecked()

    def read_type_changed(self):
        """If the read-all checkbox is checked, disables the table, if
        not, enables it.

        Parameters
        ----------

        Returns
        -------

        """
        read_all = self.checkBox_read_all.isChecked()
        if hasattr(self, "read_table"):
            self.read_table.setEnabled(not read_all)
        # if read_all:
        #     self.tableWidget_channels.setEnabled(False)
        #     self.checkBox_use_set.setEnabled(True)
        # else:
        #     self.tableWidget_channels.setEnabled(True)
        #     self.checkBox_use_set.setEnabled(False)
        self.loop_step.read_all = read_all
        self.loop_step.update_used_devices()

    def load_data(self):
        """Putting the data from the loop_step into the widgets."""
        self.checkBox_read_all.setChecked(self.loop_step.read_all)
        self.checkBox_read_variables.setChecked(self.loop_step.read_variables)
        self.checkBox_split_trigger.setChecked(self.loop_step.split_trigger)
        # self.checkBox_save.setChecked(self.loop_step.save_data)
        # self.checkBox_plot.setChecked(self.loop_step.plot_data)
        # self.checkBox_use_set.setChecked(self.loop_step.use_set_val)
        # self.build_channels_table()

    def update_step_config(self):
        """ """
        self.loop_step.channel_list = self.read_table.get_info()["channel"]
        self.loop_step.read_variables = self.checkBox_read_variables.isChecked()
        # self.lineEdit_search.clear()
        # self.build_channels_table()
        # self.loop_step.channel_list = []
        # for i in range(self.tableWidget_channels.rowCount()):
        #     if self.tableWidget_channels.item(i, 0).checkState() > 0:
        #         name = self.tableWidget_channels.item(i, 1).text()
        #         self.loop_step.channel_list.append(name)

    def table_check_changed(self, pos):
        """If a checkbox inside the table is clicked, the value is
        stored into the loopstep.

        Parameters
        ----------
        pos :


        Returns
        -------

        """
        r = pos.row()
        c = pos.column()
        if c == 0:
            name = self.tableWidget_channels.item(r, 1).text()
            if (
                self.tableWidget_channels.item(r, c).checkState()
                != Qt.CheckState.Unchecked
            ):
                self.loop_step.channel_list.append(name)
                self.loop_step.channel_list = list(set(self.loop_step.channel_list))
            elif name in self.loop_step.channel_list:
                self.loop_step.channel_list.remove(name)
            self.loop_step.update_used_devices()
        #     self.loop_step.channel_dict[name]['read'] =
        #     self.loop_step.update_used_devices()
        # if c == 2 and variables_handling.channels[name].output:
        #     self.loop_step.channel_dict[name]['use set'] = self.tableWidget_channels.item(r, c).checkState() > 0

    def build_channels_table(self):
        """This creates the table for all channels."""
        self.tableWidget_channels.clear()
        self.tableWidget_channels.setColumnCount(2)
        # self.tableWidget_channels.setRowCount(len(variables_handling.channels))
        self.tableWidget_channels.setRowCount(0)
        self.tableWidget_channels.setHorizontalHeaderLabels(["read", "channel name"])
        searchtext = self.lineEdit_search.text()
        n = 0
        for i, channel in enumerate(
            sorted(variables_handling.channels, key=lambda x: x.lower())
        ):
            if searchtext not in channel:
                continue
            self.tableWidget_channels.setRowCount(n + 1)
            item = QTableWidgetItem()
            item.setFlags(Qt.ItemIsUserCheckable | Qt.ItemIsEnabled)
            if channel in self.loop_step.channel_list:
                item.setCheckState(Qt.CheckState.Checked)
            else:
                item.setCheckState(Qt.CheckState.Unchecked)
            self.tableWidget_channels.setItem(n, 0, item)
            item = QTableWidgetItem(channel)
            item.setFlags(item.flags() ^ Qt.ItemIsEditable)
            self.tableWidget_channels.setItem(n, 1, item)
            n += 1
            # if variables_handling.channels[channel].output:
            #     item = QTableWidgetItem()
            #     item.setFlags(Qt.ItemIsUserCheckable | Qt.ItemIsEnabled)
            #     if channel in self.loop_step.channel_dict:
            #         item.setCheckState(2 if self.loop_step.channel_dict[channel]['use set'] else False)
            #     else:
            #         item.setCheckState(False)
            # else:
            #     item = QTableWidgetItem()
            #     item.setFlags(item.flags() ^ Qt.ItemIsEditable)
            # self.tableWidget_channels.setItem(i, 2, item)
        self.tableWidget_channels.resizeColumnsToContents()
        # for channel in self.loop_step.channel_dict:
        #     if channel not in variables_handling.channels:
        #         self.loop_step.channel_dict.pop(channel)


class Trigger_Channels_Step(Loop_Step):
    """
    This step provides a split between triggering and reading channels.

    Attributes
    ----------
    read_step : str
        The name of the Read Channels step, for which this step should do the
        triggering.
    """

    def __init__(self, name="", parent_step=None, step_info=None, **kwargs):
        super().__init__(name, parent_step, step_info, **kwargs)
        self.step_type = "Trigger Channels"
        if step_info is None:
            step_info = {}
        self.read_step = step_info["read_step"] if "read_step" in step_info else ""

    def get_protocol_string(self, n_tabs=1):
        """In the protocol, at first a list `channels` is defined,
        including all the channels, that are selected to be read. Then
        these channels are triggered with `helper_functions.trigger_multi`."""
        tabs = "\t" * n_tabs
        protocol_string = super().get_protocol_string(n_tabs)
        if self.read_step not in variables_handling.current_protocol.loop_step_dict:
            raise Exception(
                f'Trying to trigger channels for read_channels "{self.read_step}" but it is not there.\n{self.full_name}'
            )
        read_step = variables_handling.current_protocol.loop_step_dict[self.read_step]
        protocol_string += read_step.get_channels_string(tabs)
        step_name = read_step.variable_name()
        protocol_string += f'{tabs}grp_{step_name} = bps._short_uid("trigger")\n'
        protocol_string += f"{tabs}yield from helper_functions.trigger_multi(channels_{step_name}, grp_{step_name})\n"
        return protocol_string

    def get_protocol_short_string(self, n_tabs=0):
        """The corresponding read step is displayed."""
        short_string = super().get_protocol_short_string(n_tabs)
        read_step = variables_handling.current_protocol.loop_step_dict[self.read_step]
        short_string = f"{short_string[:-1]} - {read_step.channel_list}\n"
        return short_string


class Trigger_Channels_Config(Loop_Step_Config):
    """ """

    def __init__(self, loop_step: Trigger_Channels_Step, parent=None):
        super().__init__(parent, loop_step)
        label = QLabel("Corresponding Read-Step:")
        self.comboBox_read_step = QComboBox()
        self.layout().addWidget(label, 1, 0)
        self.layout().addWidget(self.comboBox_read_step, 1, 1, 1, 4)
        triggerable = []
        for name, step in variables_handling.current_protocol.loop_step_dict.items():
            if step.step_type == "Read Channels" and step.split_trigger:
                triggerable.append(name)
        self.comboBox_read_step.addItems(triggerable)

    def update_step_config(self):
        """ """
        self.loop_step.read_step = self.comboBox_read_step.currentText()
        super().update_step_config()



from PySide6.QtWidgets import (
    QPushButton,
    QApplication,
    QMainWindow,
    QLabel,
    QStyle,
    QFrame,
    QMenu,
    QMessageBox,
)
from PySide6.QtGui import QPainter, QColor, QBrush, QAction, QDrag
from PySide6.QtCore import Qt, QSize, QMimeData, Signal, SIGNAL

from nomad_camels.utility.variables_handling import get_color


class SimpleWrapLabel(QLabel):
    """ """

    def __init__(self, text="", parent=None):
        if " " not in text:
            text = text.replace("_", " ")
            text = text.replace("-", " ")
        super().__init__(text=text, parent=parent)

    def setText(self, a0: str) -> None:
        """

        Parameters
        ----------
        a0: str :


        Returns
        -------

        """
        if " " not in a0:
            a0 = a0.replace("_", " ")
            a0 = a0.replace("-", " ")
        super().setText(a0)


class Dots_Button(QPushButton):
    """ """

    def __init__(self, parent=None, size=100):
        super().__init__(parent)
        self.setFixedSize(size, size)
        self.setIconSize(self.size())
        self.setFlat(True)
        self.setCursor(Qt.PointingHandCursor)
        # self.setStyleSheet("QPushButton {text-align: left; font-weight: bold; font-size: 10pt; padding-bottom: -60px; font-family: Calibri; }")

    def paintEvent(self, event):
        """

        Parameters
        ----------
        event :


        Returns
        -------

        """
        painter = QPainter(self)
        painter.setRenderHint(QPainter.Antialiasing)

        # Set the brush and pen for the painter
        brush = QBrush(QColor(get_color("black")))
        painter.setBrush(brush)
        painter.setPen(Qt.NoPen)

        # Draw three dots using the painter
        size = self.size().width()
        painter.drawEllipse(size - 10, size - 20, 4, 4)
        painter.drawEllipse(size - 10, size - 13, 4, 4)
        painter.drawEllipse(size - 10, size - 6, 4, 4)
        super().paintEvent(event)

    def sizeHint(self):
        """ """
        return self.minimumSizeHint()


class Options_Run_Button(QFrame):
    """ """

    build_asked = Signal()
    external_asked = Signal()
    data_path_asked = Signal()
    del_asked = Signal()
    move_asked = Signal()

    def __init__(self, text="", size=120, small_text="run", protocol_options=True):
        super().__init__()
        self.label = SimpleWrapLabel(text, parent=self)
        self.label.setGeometry(5, 0, size - 20, size)
        self.label.setWordWrap(True)
        self.label.setStyleSheet(
            "QLabel {font-weight: bold; font-size: 11pt; font-family: Calibri;}"
        )
        self.label.setAlignment(Qt.AlignLeft | Qt.AlignBottom)
        self.button = Dots_Button(parent=self, size=size)
        self.button.setGeometry(0, 0, size, size)
        self.small_button = QPushButton(small_text, parent=self)
        icon = self.style().standardIcon(QStyle.StandardPixmap.SP_MediaPlay)
        self.small_button.setIcon(icon)
        if protocol_options:
            self.small_button.setIconSize(QSize(int(size / 5), int(size / 5)))
            self.small_button.setGeometry(5, 5, size - 10, int(size / 4))
            self.queue_button = QPushButton("queue", parent=self)
            icon = self.style().standardIcon(QStyle.StandardPixmap.SP_CommandLink)
            self.queue_button.setIcon(icon)
            self.queue_button.setIconSize(QSize(int(size / 5), int(size / 5)))
            self.queue_button.setGeometry(5, 10 + size // 4, size - 10, int(size / 4))
            self.button.setToolTip(
                "open the protocol settings / measurement sequence\nright-click for more options"
            )
            self.small_button.setToolTip("run this protocol right away")
            self.queue_button.setToolTip(
                "queue this protocol\nyou can set the protocol's variables for this specific run"
            )
        else:
            self.small_button.setIconSize(QSize(int(size / 4), int(size / 4)))
            self.small_button.setGeometry(5, 5, size - 10, int(size / 3))
            self.queue_button = None
            self.button.setToolTip("open the settings of the manual control\nright-click for more options")
            self.small_button.setToolTip("start the manual control")
        self.setContextMenuPolicy(Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(self.options_menu)
        self.config_function = None
        self.run_function = None
        self.build_function = None
        self.external_function = None
        self.data_path_function = None
        self.del_function = None
        self.move_function = None
        self.queue_function = None

        self.setFixedSize(size, size)
        self.setFrameStyle(1)
        self.protocol_options = protocol_options

    def update_functions(self):
        """ """
        # check if build_asked has anything to disconnect
        if self.receivers(SIGNAL("build_asked()")) > 0:
            self.build_asked.disconnect()
        if self.receivers(SIGNAL("external_asked()")) > 0:
            self.external_asked.disconnect()
        if self.receivers(SIGNAL("data_path_asked()")) > 0:
            self.data_path_asked.disconnect()
        if self.receivers(SIGNAL("del_asked()")) > 0:
            self.del_asked.disconnect()
        if self.receivers(SIGNAL("move_asked()")) > 0:
            self.move_asked.disconnect()
        if self.config_function is not None:
            self.button.clicked.connect(self.config_function)
        if self.run_function is not None:
            self.small_button.clicked.connect(self.run_function)
        if self.build_function is not None:
            self.build_asked.connect(self.build_function)
        if self.external_function is not None:
            self.external_asked.connect(self.external_function)
        if self.data_path_function is not None:
            self.data_path_asked.connect(self.data_path_function)
        if self.del_function is not None:
            self.del_asked.connect(self.del_function)
        if self.move_function is not None:
            self.move_asked.connect(self.move_function)
        if self.queue_function is not None:
            self.queue_button.clicked.connect(self.queue_function)

    def options_menu(self, pos):
        """

        Parameters
        ----------
        pos :


        Returns
        -------

        """
        menu = QMenu()
        actions = []
        if self.protocol_options:
            action_export = QAction("Export Protocol")
            action_export.triggered.connect(self.build_asked.emit)
            action_open = QAction("Open Externally")
            action_open.triggered.connect(self.external_asked.emit)
            action_datapath = QAction("Open Data Path")
            action_datapath.triggered.connect(self.data_path_asked.emit)
            actions += [action_export, action_open, action_datapath]
        action_move = QAction("Move to other Tab")
        action_move.triggered.connect(self.move_button)
        action_delete = QAction("Delete")
        action_delete.triggered.connect(self.delete_button)
        menu.addActions(actions)
        menu.addSeparator()
        menu.addActions([action_move, action_delete])
        menu.exec_(self.mapToGlobal(pos))

    def move_button(self):
        """ """
        self.move_asked.emit()

    def delete_button(self):
        """ """
        del_dialog = QMessageBox.question(
            self,
            f"Delete {self.label.text()}?",
            f"{self.label.text()} will be removed completely",
            QMessageBox.Yes | QMessageBox.No,
        )
        if del_dialog == QMessageBox.Yes:
            self.del_asked.emit()

    def mouseMoveEvent(self, event):
        """

        Parameters
        ----------
        event :


        Returns
        -------

        """
        if event.buttons() == Qt.LeftButton:
            mimeData = QMimeData()
            drag = QDrag(self)
            drag.setMimeData(mimeData)
            drag.exec_(Qt.MoveAction)

    def rename(self, new_name):
        """

        Parameters
        ----------
        new_name :


        Returns
        -------

        """
        self.label.setText(new_name)
        if self.config_function is not None:
            self.button.clicked.disconnect(self.config_function)
        if self.run_function is not None:
            self.small_button.clicked.disconnect(self.run_function)
        if self.build_function is not None:
            self.build_asked.disconnect(self.build_function)
        if self.external_function is not None:
            self.external_asked.disconnect(self.external_function)
        if self.del_function is not None:
            self.del_asked.disconnect(self.del_function)
        if self.move_function is not None:
            self.move_asked.disconnect(self.move_function)
        if self.queue_function is not None:
            self.queue_button.clicked.disconnect(self.queue_function)


if __name__ == "__main__":
    app = QApplication([])
    widge = QMainWindow()
    widget = Options_Run_Button("super mega test")
    # widge.setLayout(QHBoxLayout())
    # button = Options_Run_Button('s;aiouf;lwkeja;slkjfasd;lje')
    # widge.layout().addWidget(button)
    widget.show()
    app.exec()



"""This package provides access to all classes of Manual_Control that
can be used by CAMELS, including device specific ones."""

from PySide6.QtWidgets import QDialog, QGridLayout, QDialogButtonBox, QComboBox, QLabel
from PySide6.QtCore import Qt

from nomad_camels.utility import variables_handling
from nomad_camels.manual_controls.stage_control import stage_control
from nomad_camels.manual_controls.set_panel import set_panel

manual_controls = {
    "Stage_Control": [stage_control.Stage_Control, stage_control.Stage_Control_Config],
    "Set_Panel": [set_panel.Set_Panel, set_panel.Set_Panel_Config],
}


class New_Manual_Control_Dialog(QDialog):
    """A dialog that provides the user with a selection of available manual
    controls."""

    def __init__(self, parent=None):
        super().__init__(parent=parent)
        layout = QGridLayout()
        self.setLayout(layout)

        self.setWindowTitle("New Manual Control - NOMAD CAMELS")

        self.buttonBox = QDialogButtonBox()
        self.buttonBox.setOrientation(Qt.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.Cancel | QDialogButtonBox.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.buttonBox.rejected.connect(self.reject)
        self.buttonBox.accepted.connect(self.accept)

        self.selection_box = QComboBox()
        self.std_controls = manual_controls
        self.selection_box.addItems(
            sorted(self.std_controls.keys(), key=lambda x: x.lower())
        )
        self.instr_controls = get_instrument_controls()
        self.selection_box.addItems(
            sorted(self.instr_controls.keys(), key=lambda x: x.lower())
        )

        self.selected_control = None

        label = QLabel("What kind of manual control do you want to add?")
        layout.addWidget(label, 0, 0)
        layout.addWidget(self.selection_box, 1, 0)
        layout.addWidget(self.buttonBox, 2, 0)

    def accept(self):
        """Writes the control that was selected into `self.selected_control`."""
        name = self.selection_box.currentText()
        if name in self.std_controls:
            self.selected_control = self.std_controls[name]
        else:
            self.selected_control = self.instr_controls[name]
        super().accept()


def get_instrument_controls():
    """Goes through all configured instruments and checks them for manual
    controls. If there are any, they will be added to a dictionary.

    Returns
    -------
    controls : dict{'<name>': (Manual_Control, Manual_Control_Config)}
        Dictionary with the manual controls and there config widgets.
    """
    controls = {}
    for name, instr in sorted(
        variables_handling.devices.items(), key=lambda x: x[0].lower()
    ):
        adds = instr.get_controls()
        controls.update(adds)
    return controls


def get_control_by_type_name(name):
    """
    Returns a tuple/list of the specified control's class / config-class.

    Parameters
    ----------
    name : str
        Name of the manual control (could be CAMELS-main or instrument specific)

    Returns
    -------
    tuple(Manual_Control, Manual_Control_Config)
    """
    if name in manual_controls:
        return manual_controls[name]
    instr_controls = get_instrument_controls()
    if name in instr_controls:
        return instr_controls[name]
    raise Exception(f"Manual Control of type {name} is not defined!")



from tqdm import tqdm


class ProgressBar:
    """Provides a progress bar for protocols run directly from the script.
    With the `emit` method, the bar can be updated the same way signals are
    used when running the protocol from the UI.

    Attributes
    ----------
    n_steps : int
        the number of total (expected) steps of the protocol
    bar : tqdm
        the actual progress bar
    """

    def __init__(self, n_steps):
        self.n_steps = n_steps
        self.bar = tqdm(total=n_steps)

    def emit(self, n):
        """
        Updates the progress bar to the next step. The parameter `n` is only so
        the function has the same shape as the `emit` function for the signal
        used with the main UI.

        Parameters
        ----------
        n : int
            has no function, only there to be able to give this method a value
        """
        self.bar.update()



"""
This module provides functions that are often used when building a protocol.
It is not inside the `protocol_builder` module, since these functions are also
called by the classes of the single protocol-steps.
"""

import copy

standard_plot_string = "\tglobal app\n"
standard_plot_string += "\tapp = QCoreApplication.instance()\n"
standard_plot_string += "\tif app is None:\n"
standard_plot_string += "\t\tapp = QApplication(sys.argv)\n"
# standard_plot_string += '\tapp.aboutToQuit.connect(wait_for_workers_to_quit)\n'
standard_plot_string += "\tif darkmode:\n"
standard_plot_string += "\t\tplot_pyqtgraph.activate_dark_mode()\n"
# standard_plot_string += '\ttheme_changing.change_theme(theme, app)\n'
# standard_plot_string += '\t\timport qdarkstyle\n'
# standard_plot_string += '\t\tapp.setStyleSheet(qdarkstyle.load_stylesheet_pyqt5())\n'
# standard_plot_string += '\t\tapp.setStyleSheet(qdarkstyle.load_stylesheet(qt_api="pyqt5"))\n'


def get_plot_add_string(name, stream, subprotocol=False, n_tabs=1):
    """
    This function generates the necessary lines for the "steps_add_main"
    function in the protocol-script to open the plots of a sub-step of the
    protocol.

    Parameters
    ----------
    name : str
        The name of the calling protocol-step.

    stream : str
        The bluesky-stream on which the plots should react.

    subprotocol :
         (Default value = False)
         If True, the create_plots function will be called from the
         subprotocol's module.

    Returns
    -------
    add_main_string
        The string to be added.

    """
    tabs = "\t" * n_tabs
    add_main_string = f'{tabs}if "subs" not in returner:\n'
    add_main_string += f'{tabs}\treturner["subs"] = []\n'
    add_main_string += f'{tabs}if "plots" not in returner:\n'
    add_main_string += f'{tabs}\treturner["plots"] = []\n'
    if subprotocol:
        add_main_string += (
            f"{tabs}plots, subs, _ = {name}_mod.create_plots(RE, {stream})\n"
        )
    else:
        add_main_string += f"{tabs}plots, subs, _ = create_plots_{name}(RE, {stream})\n"
    add_main_string += f'{tabs}returner["subs"] += subs\n'
    add_main_string += f'{tabs}returner["plots"] += plots\n'
    return add_main_string


def plot_creator(
    plot_data,
    func_name="create_plots",
    multi_stream=False,
    plot_is_box=False,
    box_names="",
    skip_box=True,
):
    """
    Creates the `create_plots` function for the protocol.

    Parameters
    ----------
    plot_data : list[Plot_Info]
        A list of Plot_Info objects with all information for the plots.

    func_name : str
         (Default value = 'create_plots')
         The name of the function in the protocol-script.

    multi_stream :
         (Default value = False)
         Passes the multi_stream keyword to the plot-widgets. This is used if
         the plot should not only react to the stream that's specified exactly.

    Returns
    -------
    plot_string
        The string containing the function for the protocol-script.

    plotting
        True if there actually was a plot. Otherwise the function will not be
        created in the script at all.


    """
    plot_string = f'\ndef {func_name}(RE, stream="primary"):\n'
    if not plot_data:
        plot_string += "\treturn [], [], None\n\n"
        return plot_string, False
    plot_string += standard_plot_string
    plot_string += "\tsubs = []\n"
    plotting = False
    for i, plot in enumerate(plot_data):
        if plot.plt_type == "X-Y plot":
            plotting = True
            fits = []
            if plot.same_fit:
                if plot.all_fit and plot.all_fit.do_fit:
                    for y in plot.y_axes["formula"]:
                        fit = copy.deepcopy(plot.all_fit)
                        fit.y = y
                        fits.append(fit)
            else:
                for fit in plot.fits:
                    if fit.do_fit:
                        fits.append(fit)
            plot_string += "\tfits = []\n"
            for fit in fits:
                plot_string += f"\tfits.append({fit.__dict__})\n"
            y_axes = {}
            ylabel2 = plot.ylabel2 if plot.ylabel2 else ""
            for j, f in enumerate(plot.y_axes["formula"]):
                y_axes[f] = 2 if plot.y_axes["axis"][j] == "right" else 1
                if not ylabel2 and y_axes[f] == 2:
                    ylabel2 = f
            xlabel = plot.xlabel if plot.xlabel else plot.x_axis or "time"
            ylabel = plot.ylabel if plot.ylabel else plot.y_axes["formula"][0]
            plot_string += f'\tplot_info = dict(x_name="{plot.x_axis or "time"}", y_names={plot.y_axes["formula"]}, ylabel="{ylabel}", xlabel="{xlabel}", title="{plot.title}", stream_name=stream, namespace=namespace, fits=fits, multi_stream={multi_stream}, y_axes={y_axes}, ylabel2="{ylabel2}", logX={plot.logX}, logY={plot.logY}, logY2={plot.logY2}, maxlen="{plot.maxlen}")\n'
            plot_string += f"\tplot_{i} = plot_pyqtgraph.PlotWidget(**plot_info)\n"
            plot_string += f"\tplots.append(plot_{i})\n"
            # plot_string += f"\tplot_data.append(plot_info)\n"
            plot_string += f"\tplot_{i}.show()\n"
            plot_string += f"\tsubs.append(RE.subscribe(plot_{i}.livePlot))\n"
            plot_string += f"\tfor fit in plot_{i}.liveFits:\n"
            plot_string += "\t\tall_fits[fit.name] = fit\n"
            # plot_string += f'\tfor lfp in plot_{i}.liveFitPlots:\n'
            # plot_string += f'\t\tsubs.append(RE.subscribe(lfp))\n'
        elif plot.plt_type == "Value-List":
            plotting = True
            plot_string += f'\tplot_{i} = list_plot.Values_List_Plot({plot.y_axes["formula"]}, title="{plot.title}", stream_name=stream, namespace=namespace, plot_all_available={plot.plot_all_available})\n'
            plot_string += f"\tplots.append(plot_{i})\n"
            plot_string += f"\tplot_{i}.show()\n"
            plot_string += f"\tsubs.append(RE.subscribe(plot_{i}.livePlot))\n"
        elif plot.plt_type == "2D plot":
            plotting = True
            plot_string += f'\tplot_{i} = plot_pyqtgraph.PlotWidget_2D("{plot.x_axis}", "{plot.y_axes["formula"][0]}", "{plot.z_axis}", xlabel="{plot.xlabel}", ylabel="{plot.ylabel}", zlabel="{plot.zlabel}", title="{plot.title}", maxlen="{plot.maxlen}", stream_name=stream, namespace=namespace)\n'
            plot_string += f"\tplots.append(plot_{i})\n"
            plot_string += f"\tplot_{i}.show()\n"
            plot_string += f"\tsubs.append(RE.subscribe(plot_{i}.livePlot))\n"
        if plot_is_box:
            plot_string += f"\tboxes['{box_names}_{i}'] = helper_functions.Waiting_Bar(title='{plot.title}', skipable={skip_box}, display_bar=False, plot=plot_{i})\n"
    plot_string += "\treturn plots, subs, app\n\n"
    return plot_string, plotting



from PySide6.QtWidgets import QWidget, QCheckBox, QTextEdit, QMessageBox, QPushButton
from PySide6.QtCore import Signal

from nomad_camels.frontpanels.plot_definer import Plot_Button_Overview
from nomad_camels.loop_steps import make_step_of_type
from nomad_camels.gui.general_protocol_settings import Ui_Protocol_Settings

from nomad_camels.ui_widgets.add_remove_table import AddRemoveTable
from nomad_camels.ui_widgets.path_button_edit import Path_Button_Edit
from nomad_camels.utility import variables_handling


class Measurement_Protocol:
    """Class for the measurement protocols. It mainly contains
    loop_steps and plots.

    Attributes
    ----------
    description : str
        A string describing the protocol.
    export_csv : bool
        If True, the data will be exported to a csv file in the end.
    export_json : bool
        If True, the metadata will be exported to a json file in the end.
    session_name : str
        This name is appended to the entry in the hdf5-file, or used for the
        filenames of csv / json to make it easier for the user to recognize
        their measurements
    loop_steps : list[Loop_Step]
        A list of the steps performed by the protocol. This list also represents
        the order, in which the steps should be performed. The step's children
        steps are not included in this list.
    loop_step_dict : dict
        Keys are the steps' names with the steps being the value. The sub-steps
        are included here.
    plots : list[Plot_Info]
        Contains the information for the protocol's plots.
    filename : str
        The name of the produced datafile.
    variables : dict
        Name-value pairs of the protocol variables
    loop_step_variables : dict
        Name-value pairs of the variables provided by the steps (e.g. for loop)
    channels : dict
        Dictionary of the channel-names and channels used inside the protocol
    name : str
        The name of the protocol. This also appears in the main UI
    use_nexus : bool
        Whether to write a NeXus-entry into the datafile as well.
    """

    def __init__(
        self,
        loop_steps=None,
        plots=None,
        channels=None,
        name="",
        use_nexus=True,
        **kwargs,
    ):
        if plots is None:
            plots = []
        if loop_steps is None:
            loop_steps = []
        if channels is None:
            channels = {}
        self.description = kwargs["description"] if "description" in kwargs else ""
        self.export_csv = kwargs["export_csv"] if "export_csv" in kwargs else False
        self.export_json = kwargs["export_json"] if "export_json" in kwargs else False
        self.session_name = kwargs["session_name"] if "session_name" in kwargs else ""
        self.skip_config = kwargs["skip_config"] if "skip_config" in kwargs else False
        self.h5_during_run = (
            kwargs["h5_during_run"] if "h5_during_run" in kwargs else True
        )
        self.use_end_protocol = (
            kwargs["use_end_protocol"] if "use_end_protocol" in kwargs else False
        )
        self.end_protocol = kwargs["end_protocol"] if "end_protocol" in kwargs else ""
        self.live_variable_update = (
            kwargs["live_variable_update"]
            if "live_variable_update" in kwargs
            else False
        )
        self.allow_live_comments = (
            kwargs["allow_live_comments"] if "allow_live_comments" in kwargs else False
        )
        self.loop_steps = loop_steps
        self.loop_step_dict = {}
        for step in self.loop_steps:
            update_all_children(self.loop_step_dict, step)
        self.plots = plots
        self.filename = ""
        self.variables = {}
        self.loop_step_variables = {}
        self.channels = channels
        self.name = name or "Protocol"
        self.use_nexus = use_nexus
        self.measurement_description = ""
        self.tags = []

    def update_variables(self):
        """Update all the variables provided by loopsteps."""
        self.loop_step_variables.clear()
        for step in self.loop_steps:
            if not step.is_active:
                continue
            step.update_variables()

    def add_loop_step(self, loop_step, position=-1, parent_step_name=None, model=None):
        """Adds a loop_step to the protocol (or the parent_step)at the specified
        position. Also appends the loop_step to the given model. The
        loop_step is added to the list as well as the dictionary.

        Parameters
        ----------
        loop_step : Loop_Step
            The step that should be added
        position : int
            (Default value = -1)
            Where in the list to add the step
        parent_step_name : str, None
            (Default value = None)
            If the step is not in the outermost layer, its parent's name. Then
            `position` will not be for the protocol's list, but the parent's
            children-list.
        model : QAbstractItemModel, None
            (Default value = None)
            The item model that is used to display the steps.
        """
        if parent_step_name is None:
            if position < 0:
                self.loop_steps.append(loop_step)
            else:
                self.loop_steps.insert(position, loop_step)
        else:
            loop_step.parent_step = parent_step_name
            self.loop_step_dict[parent_step_name].add_child(loop_step, position)
        if model is not None:
            loop_step.append_to_model(model)
        self.loop_step_dict.update({loop_step.full_name: loop_step})

    def add_loop_step_rec(
        self, loop_step, model=None, parent_step_name=None, position=-1
    ):
        """Recursively adds the loop_step and all its children to the
        protocol. Steps are added to the list if they have no parent,
        otherwise to the parent. All are added to the dictionary.

        Parameters
        ----------
        loop_step : Loop_Step
            The step that should be added
        model : QAbstractItemModel, None
            (Default value = None)
            The item model that is used to display the steps.
        parent_step_name : str
            (Default value = None)
            If the step is not in the outermost layer, its parent's name. Then
            `position` will not be for the protocol's list, but the parent's
            children-list.
        position : int
            (Default value = -1)
            Where in the list to add the step
        """
        if parent_step_name is None:
            self.add_loop_step(
                loop_step,
                model=model,
                parent_step_name=parent_step_name,
                position=position,
            )
        else:
            if model is not None:
                loop_step.append_to_model(model, parent=parent_step_name)
            if loop_step not in self.loop_step_dict[parent_step_name].children:
                self.loop_step_dict[parent_step_name].add_child(loop_step, position)
            self.loop_step_dict.update({loop_step.full_name: loop_step})
        for child in loop_step.children:
            self.add_loop_step_rec(
                child, parent_step_name=loop_step.full_name, model=model
            )

    def remove_loop_step(self, loop_step_name):
        """Removes the step with the given name from the sequence-list
        (or parent) and from the dictionary.

        Parameters
        ----------
        loop_step_name : str
            The name of the step that is to be removed.
        """
        step = self.loop_step_dict.pop(loop_step_name)
        if step.parent_step is not None:
            self.loop_step_dict[step.parent_step].remove_child(step)
        else:
            self.loop_steps.remove(step)

    def load_loop_steps(self, loop_steps, model=None):
        """Takes a list of loop_steps, creates them (with the input data
        of each step) and adds them to the specified model.

        Parameters
        ----------
        loop_steps : list
            A list containing all the information (as dictionary) for creating
            the single steps.
        model : QAbstractItemModel, None
            (Default value = None)
            The item model that is used to display the steps.
        """
        for step in loop_steps:
            loop_step = self.make_step(step)
            self.add_loop_step_rec(loop_step, model=model)

    def make_step(self, step_info):
        """Creates the step specified with step_info (including the
        children), 'step_type' gives which subclass of Loop_Step shall
        be created.

        Parameters
        ----------
        step_info : dict
            This dictionary should hold all information needed to create the
            step. Specifically there should be "step_type" to decide on the
            class of step and "full_name" to give the step its name.


        Returns
        -------
        st : Loop_Step
            The created step
        """
        # children = None
        # if step_info['has_children']:
        children = []
        if "children" in step_info:
            for child in step_info["children"]:
                child_step = self.make_step(child)
                child_step.parent_step = step_info["full_name"]
                children.append(child_step)
        st = make_step_of_type.make_step(
            step_info["step_type"], step_info, children, protocol=self
        )
        st.full_name = step_info["full_name"]
        return st

    def rearrange_loop_steps(self, step_list):
        """Takes a list of loopsteps, each entry consisting of a tuple
        of the loopstep name and its children, which is recursively the
        same kind of list. Re-populates the loop_step_dict and then puts
        the loop_steps in the correct order.

        Parameters
        ----------
        step_list : list[(str, [list(str, ...)])]
            Contains tuples with the first entry being the names of the steps in
            the order they should be arranged to and the second entry being a
            list of the steps' children. That list should be structured the same
            way as this `step_list`.
        """
        self.loop_step_dict = {}
        for step in self.loop_steps:
            update_all_children(self.loop_step_dict, step)
        self.loop_steps = []
        for step, children in step_list:
            self.loop_step_dict[step].children = []
            append_all_children(
                children, self.loop_step_dict[step], self.loop_step_dict
            )
            self.loop_steps.append(self.loop_step_dict[step])

    def get_plan_string(self):
        """Get the string for the protocol-plan, including the loopsteps."""
        variables_handling.current_protocol = self
        plan_string = f'\n\n\ndef {self.name.replace(" ","_")}_plan_inner(devs, eva=None, stream_name="primary", runEngine=None):\n'
        prot_vars = dict(variables_handling.protocol_variables)
        if "StartTime" in prot_vars:
            prot_vars.pop("StartTime")
            prot_vars.pop("ElapsedTime")
        if prot_vars:
            plan_string += "\tglobal "
            for i, var in enumerate(prot_vars.keys()):
                if i > 0:
                    plan_string += ", "
                plan_string += var
            plan_string += "\n"
        for step in self.loop_steps:
            if not step.is_active:
                continue
            plan_string += step.get_protocol_string(n_tabs=1)
        plan_string += f'\n\n\ndef {self.name.replace(" ","_")}_plan(devs, md=None, runEngine=None, stream_name="primary"):\n'
        plan_string += "\teva = Evaluator(namespace=namespace)\n"
        plan_string += "\tsub_eva = runEngine.subscribe(eva)\n"
        plan_string += "\tyield from bps.open_run(md=md)\n"
        if self.use_end_protocol:
            plan_string += "\ttry:\n"
            plan_string += f'\t\tyield from {self.name.replace(" ", "_")}_plan_inner(devs, eva, stream_name, runEngine)\n'
            plan_string += "\t\tyield from helper_functions.get_fit_results(all_fits, namespace, True)\n"
            plan_string += "\tfinally:\n"
            plan_string += "\t\tyield from ending_steps(runEngine, devs)\n"
        else:
            plan_string += f'\tyield from {self.name.replace(" ", "_")}_plan_inner(devs, eva, stream_name, runEngine)\n'
            plan_string += "\tyield from helper_functions.get_fit_results(all_fits, namespace, True)\n"
        check_live_windows = False
        if self.allow_live_comments:
            check_live_windows = True
        if check_live_windows:
            plan_string += "\tfinished = False\n"
            plan_string += "\twhile not finished:\n"
            plan_string += "\t\tfinished = True\n"
            plan_string += "\t\tfor window in live_windows:\n"
            plan_string += "\t\t\tif hasattr(window, '_is_finished') and not window._is_finished:\n"
            plan_string += "\t\t\t\tfinished = False\n"
            plan_string += "\tlive_metadata = {}\n"
            plan_string += "\tfor window in live_windows:\n"
            plan_string += "\t\tif hasattr(window, 'get_metadata'):\n"
            plan_string += "\t\t\tlive_metadata.update(window.get_metadata())\n"
            plan_string += "\tlive_metadata_signal = variable_reading.Variable_Signal(name='live_metadata', variables_dict=live_metadata)\n"
            plan_string += '\tyield from bps.trigger_and_read([live_metadata_signal], name="_live_metadata_reading_")\n'
        plan_string += "\tyield from bps.close_run()\n"
        plan_string += "\trunEngine.unsubscribe(sub_eva)\n"
        return plan_string

    def get_short_string(self):
        """Goes through all steps and creates an overview of what is happening
        in the protocol."""
        short_string = ""
        for step in self.loop_steps:
            if not step.is_active:
                continue
            short_string += step.get_protocol_short_string()
        return short_string

    def get_add_main_string(self):
        """Gets all the steps that should be executed in the protocol's main
        function."""
        add_main_string = "def steps_add_main(RE, devs):\n"
        add_main_string += "\treturner = {}\n"
        for step in self.loop_steps:
            if not step.is_active:
                continue
            add_main_string += step.get_add_main_string()
        if self.use_end_protocol:
            from nomad_camels.bluesky_handling import protocol_builder

            add_main_string += protocol_builder.make_plots_string_of_protocol(
                self.end_protocol, 1
            )

        add_main_string += "\treturn returner\n\n\n"
        return add_main_string

    def get_live_interaction_string(self):
        """Returns the string for the live interaction of the protocol."""
        live_string = "def create_live_windows():\n"
        live_string += "\tglobal live_windows\n"
        if self.live_variable_update:
            live_string += (
                "\tfrom nomad_camels.ui_widgets.variable_table import VariableBox\n"
            )
            live_string += f"\tvariables = {self.variables}\n"
            live_string += f'\tvariable_box = VariableBox(editable_names=False, variables=variables, name="{self.name}")\n'
            live_string += "\n\tdef update_variables(new_variables):\n"
            live_string += "\t\tglobal namespace\n"
            live_string += "\t\tnamespace.update(new_variables)\n\n"
            live_string += (
                "\tvariable_box.new_values_signal.connect(update_variables)\n"
            )
            live_string += "\tlive_windows.append(variable_box)\n"
            live_string += "\tvariable_box.show()\n"
        if self.allow_live_comments:
            live_string += "\tcommenting_box = helper_functions.Commenting_Box()\n"
            live_string += "\tlive_windows.append(commenting_box)\n"
        live_string += "\treturn live_windows\n\n\n"
        return live_string

    def get_total_steps(self):
        """Returns the total number of steps (including repetitions for loops)"""
        total = 0
        for step in self.loop_steps:
            if not step.is_active:
                continue
            step.update_time_weight()
            total += step.time_weight
        return total

    def get_outer_string(self):
        """Strings outside of all other functions of the script, e.g. more
        functions to create step-specific plots."""
        outer_string = "\n"
        for step in self.loop_steps:
            if not step.is_active:
                continue
            outer_string += step.get_outer_string()
        if self.use_end_protocol:
            from nomad_camels.bluesky_handling import protocol_builder

            outer_string += protocol_builder.import_protocol_string(
                self.end_protocol, 0
            )
        return outer_string

    def get_used_devices(self):
        """Get a list of all devices needed by any loopstep."""
        devices = []
        for step in self.loop_steps:
            if not step.is_active:
                continue
            step.update_used_devices()
            devices += step.used_devices
        adds = []
        for dev in devices:
            adds += variables_handling.devices[dev].get_necessary_devices()
        devices += adds
        if self.use_end_protocol:
            from nomad_camels.utility import load_save_functions

            end_protocol = load_save_functions.load_protocol(self.end_protocol)
            devices += end_protocol.get_used_devices()
        devices = list(set(devices))
        devices = sorted(devices, key=lambda x: x in adds, reverse=True)
        return devices


def append_all_children(child_list, step, step_dict):
    """Takes a list of the kind specified in rearrange_loop_steps, does
    the same as the other function, but recursively for all the
    (grand-)children.

    Parameters
    ----------
    child_list : list[(str, [list(str, ...)])]
        Contains tuples with the first entry being the names of the steps in
        the order they should be arranged to and the second entry being a
        list of the steps' children. That list should be structured the same
        way as this `child_list`.
    step : Loop_Step_Container
        The step to which the children should be appended.
    step_dict : dict
        A dictionary containing (among others) the steps of `child_list`.
    """
    for child, grandchildren in child_list:
        child_step = step_dict[child]
        child_step.children = []
        append_all_children(grandchildren, child_step, step_dict)
        child_step.parent_step = step.full_name
        step.children.append(child_step)


def update_all_children(step_dict, step):
    """Similar to append_all_children, but only updating the step_dict
    with all the children, i.e. writing them into the dictionary.

    Parameters
    ----------
    step_dict : dict
        Dictionary, where to write all the steps
    step : Loop_Step
        The step to be added to the dictionary. If it has child-steps, all of
        them will be added to the dictionary recursively.
    """
    step_dict.update({step.full_name: step})
    # if step.has_children:
    for child in step.children:
        update_all_children(step_dict, child)


class General_Protocol_Settings(Ui_Protocol_Settings, QWidget):
    """Widget for the configuration of the general protocol settings.
    Here plots may be defined and variables added to the protocol.

    Parameters
    ----------

    Returns
    -------

    """

    name_changed = Signal()

    def __init__(self, parent=None, protocol=Measurement_Protocol()):
        super(General_Protocol_Settings, self).__init__(parent)
        self.setupUi(self)
        self.protocol = protocol
        self.lineEdit_filename.setText(self.protocol.filename)
        self.lineEdit_protocol_name.setText(self.protocol.name)

        self.variable_table.set_protocol(self.protocol)
        self.variable_table.editable_names = True

        # self.variable_model = QStandardItemModel()
        # self.variable_model.setHorizontalHeaderLabels(["Name", "Value", "Data-Type"])
        # self.tableView_variables.setModel(self.variable_model)
        # self.load_variables()

        self.pushButton_add_variable.clicked.connect(lambda x: self.add_variable())
        self.pushButton_remove_variable.clicked.connect(self.remove_variable)

        # self.variable_model.itemChanged.connect(self.check_variable)

        self.checkBox_perform_at_end = QCheckBox("Perform steps at end of protocol")
        self.checkBox_perform_at_end.setChecked(self.protocol.use_end_protocol)
        self.ending_protocol_selection = Path_Button_Edit(
            self,
            self.protocol.end_protocol,
            default_dir=variables_handling.preferences["py_files_path"],
            file_extension="*.cprot",
        )
        self.checkBox_perform_at_end.setToolTip(
            "Select a protocol to be performed at the end of this protocol or when it is aborted by the user.\nThis may be useful e.g. to turn something of in a controlled way.\nThis is NOT executed, when the protocol is run as a subprotocol."
        )
        self.ending_protocol_selection.setToolTip(
            "Select a protocol to be performed at the end of this protocol or when it is aborted by the user.\nThis may be useful e.g. to turn something of in a controlled way.\nThis is NOT executed, when the protocol is run as a subprotocol."
        )
        self.checkBox_perform_at_end.stateChanged.connect(self.check_use_ending_steps)
        self.check_use_ending_steps()

        self.plot_widge = Plot_Button_Overview(self, self.protocol.plots)

        self.checkBox_NeXus.setChecked(self.protocol.use_nexus)
        self.lineEdit_protocol_name.textChanged.connect(self.name_change)
        self.name_change()

        self.textEdit_desc_protocol = QTextEdit(parent=self)
        self.textEdit_desc_protocol.textChanged.connect(self.adjust_text_edit_size_prot)
        self.textEdit_desc_protocol.setPlaceholderText("Enter your description here.")
        if self.protocol.description:
            self.textEdit_desc_protocol.setText(self.protocol.description)
        self.checkBox_csv_exp.setChecked(self.protocol.export_csv)
        self.checkBox_json_exp.setChecked(self.protocol.export_json)
        self.checkBox_no_config.setChecked(self.protocol.skip_config)
        self.checkBox_no_config.clicked.connect(self.enable_disable_config)
        self.adjust_text_edit_size_prot()

        self.checkBox_live_variables.setChecked(self.protocol.live_variable_update)
        self.checkBox_live_comments.setChecked(self.protocol.allow_live_comments)

        self.radioButton_h5_during.setChecked(self.protocol.h5_during_run)
        self.radioButton_h5_after.setChecked(not self.protocol.h5_during_run)

        self.layout().addWidget(self.textEdit_desc_protocol, 5, 0, 1, 6)

        self.layout().addWidget(self.plot_widge, 6, 0, 1, 6)
        self.layout().addWidget(self.checkBox_perform_at_end, 20, 0, 1, 6)
        self.layout().addWidget(self.ending_protocol_selection, 21, 0, 1, 6)
        # ! Ui_Protocol_Settings.ui file adds the Variables Table at position 8 and 9 !!!

        self.update_variable_select()
        self.variable_table.selectionModel().selectionChanged.connect(
            self.update_variable_select
        )

    def check_use_ending_steps(self):
        """If the checkBox_perform_at_end is checked, the ending_protocol_selection
        is enabled, otherwise disabled."""
        if self.checkBox_perform_at_end.isChecked():
            self.ending_protocol_selection.setEnabled(True)
        else:
            self.ending_protocol_selection.setEnabled(False)

    def showEvent(self, event):
        """Called when the widget is shown."""
        super().showEvent(event)
        self.adjust_text_edit_size_prot()

    def adjust_text_edit_size_prot(self):
        """Adjusts the size of the textEdit_desc_protocol based on its content."""
        document = self.textEdit_desc_protocol.document()
        document_height = document.size().height()
        self.textEdit_desc_protocol.setFixedHeight(
            document_height + 5
        )  # Add some padding

    def enable_disable_config(self):
        disabling = self.checkBox_no_config.isChecked()
        if disabling:
            msgBox = QMessageBox()
            msgBox.setText(
                "Are you sure you want to disable configuration of the used instruments?\nThis may lead to unexpected behaviour of the instruments, if they are not configured correctly beforehand!"
            )
            msgBox.setStandardButtons(QMessageBox.Yes | QMessageBox.No)
            msgBox.setDefaultButton(QMessageBox.No)
            result = msgBox.exec()
            if result != QMessageBox.Yes:
                self.checkBox_no_config.setChecked(False)

    def add_variable(self):
        """Add a variable to the list, given a unique name, then updates
        the protocol.

        Parameters
        ----------

        Returns
        -------

        """
        self.variable_table.append_variable("name")
        self.variable_table.update_variables()
        self.update_variable_select()

    # def append_variable(self, name="name", value="value"):
    #     """Append the variable with name and value to the item_model,
    #     also add an item that shows the datatype of the value.

    #     Parameters
    #     ----------
    #     name :
    #          (Default value = 'name')
    #     value :
    #          (Default value = 'value')

    #     Returns
    #     -------

    #     """
    #     name_item = QStandardItem(name)
    #     value_item = QStandardItem(value)
    #     type_item = QStandardItem(variables_handling.check_data_type(value))
    #     type_item.setEditable(False)
    #     self.variable_model.appendRow([name_item, value_item, type_item])

    def remove_variable(self):
        """Removes the selected variable."""
        try:
            index = self.variable_table.selectedIndexes()[0]
        except IndexError:
            raise Exception("You need to select a row first!")
        if index.row() >= 0:
            self.variable_table.model.removeRow(index.row())
            self.variable_table.update_variables()

    def update_step_config(self):
        """Updates all the protocol settings."""
        self.protocol.filename = self.lineEdit_filename.text()
        self.protocol.name = self.lineEdit_protocol_name.text()
        self.protocol.description = self.textEdit_desc_protocol.toPlainText()
        self.protocol.plots = self.plot_widge.plot_data
        self.protocol.export_csv = self.checkBox_csv_exp.isChecked()
        self.protocol.export_json = self.checkBox_json_exp.isChecked()
        self.protocol.skip_config = self.checkBox_no_config.isChecked()
        self.variable_table.update_variables()
        self.protocol.use_nexus = self.checkBox_NeXus.isChecked()
        self.protocol.h5_during_run = self.radioButton_h5_during.isChecked()
        self.protocol.use_end_protocol = self.checkBox_perform_at_end.isChecked()
        self.protocol.end_protocol = self.ending_protocol_selection.get_path()
        self.protocol.live_variable_update = self.checkBox_live_variables.isChecked()
        self.protocol.allow_live_comments = self.checkBox_live_comments.isChecked()

    # def load_variables(self):
    #     """Called when starting, loads the variables from the protocol
    #     into the table.

    #     Parameters
    #     ----------

    #     Returns
    #     -------

    #     """
    #     for var in sorted(self.protocol.variables):
    #         self.append_variable(var, str(self.protocol.variables[var]))

    # def check_variable(self):
    #     """If name of variable changed: check whether the variable is
    #     unique, if not change its name and raise an error.
    #     If value changed: re-evaluate the data-type.
    #     Update the protocol afterwards.

    #     Parameters
    #     ----------

    #     Returns
    #     -------

    #     """
    #     ind = self.tableView_variables.selectedIndexes()
    #     if ind:
    #         ind = ind[0]
    #     else:
    #         return
    #     item = self.variable_model.itemFromIndex(ind)
    #     if ind.column() == 0:
    #         variables_handling.check_variable_name(item.text(), parent=self)
    #     if ind.column() == 0 and item.text() in self.protocol.variables:
    #         new_name = self.get_unique_name(item.text())
    #         item.setText(new_name)
    #         raise Exception("Variable names must be unique!")
    #     if ind.column() == 1:
    #         d_type = variables_handling.check_data_type(item.text())
    #         self.variable_model.item(ind.row(), 2).setText(d_type)
    #     self.update_variables()

    # def update_variables(self):
    #     """Taking all the variables from the list into the protocol."""
    #     self.protocol.variables = {}
    #     for i in range(self.variable_model.rowCount()):
    #         name = self.variable_model.item(i, 0).text()
    #         value = variables_handling.get_data(self.variable_model.item(i, 1).text())
    #         self.protocol.variables.update({name: value})
    #     variables_handling.protocol_variables = self.protocol.variables
    #     self.update_variable_select()

    def update_variable_select(self):
        if self.variable_table.selectedIndexes():
            self.pushButton_remove_variable.setEnabled(True)
        else:
            self.pushButton_remove_variable.setEnabled(False)

    def name_change(self):
        """ """
        name = self.lineEdit_protocol_name.text()
        self.label_title.setText(f"{name} - General Configuration")
        self.protocol.name = name
        self.name_changed.emit()



import importlib
import os.path
import sys
import databroker
import pytest

from PySide6.QtCore import Qt, QItemSelectionModel
from PySide6.QtWidgets import QMessageBox

from nomad_camels.frontpanels import protocol_config
from nomad_camels.bluesky_handling import make_catalog
from nomad_camels.frontpanels import instrument_installer
from nomad_camels.utility import variables_handling
from nomad_camels.utility.treeView_functions import getItemIndex


@pytest.fixture(autouse=True)
def mock_message_box(monkeypatch):
    """If a messagebox, e.g. asking whether to discard all changes to the
    protocol, pops up, it is automatically accepted"""

    def mock_question(*args, **kwargs):
        return QMessageBox.Yes

    monkeypatch.setattr(QMessageBox, "question", mock_question)


def test_change_dev_config(qtbot, tmp_path):
    """Opens the config for "change device config" tries to configure it for the
    demo instrument and tries to run a protocol with this step."""
    ensure_demo_in_devices()
    from nomad_camels.loop_steps import change_device_config

    conf = protocol_config.Protocol_Config()
    qtbot.addWidget(conf)
    action = get_action_from_name(conf.add_actions, "Change Device Config")
    action.trigger()
    conf_widge = conf.loop_step_configuration_widget
    assert isinstance(conf_widge, change_device_config.Change_DeviceConf_Config)
    conf_widge.comboBox_device.setCurrentText("demo_instrument")
    with qtbot.waitSignal(conf.accepted) as blocker:
        conf.accept()
    prot = conf.protocol
    prot.name = "test_change_dev_config_protocol"
    assert "Change Device Config (Change_Device_Config)" in prot.loop_step_dict
    assert prot.loop_steps[0].device == "demo_instrument"
    catalog_maker(tmp_path)
    run_test_protocol(tmp_path, prot)


def test_for_loop(qtbot, tmp_path):
    """Opens the config for a "For Loop" tries to configure it looping over a
    wait-step and tries to run a protocol with this step."""
    # ensure_demo_in_devices()
    from nomad_camels.loop_steps import for_while_loops

    conf = protocol_config.Protocol_Config()
    qtbot.addWidget(conf)
    action = get_action_from_name(conf.add_actions, "For Loop")
    action.trigger()
    conf_widge = conf.loop_step_configuration_widget
    assert isinstance(conf_widge, for_while_loops.For_Loop_Step_Config)
    conf_widge.sub_widget.lineEdit_start.setText("0")
    conf_widge.sub_widget.lineEdit_stop.setText("1")
    conf_widge.sub_widget.lineEdit_n_points.setText("11")

    action = get_action_from_name(conf.add_actions, "Wait")
    action.trigger()
    prot = conf.protocol

    select_step_by_name(conf, "Wait (Wait)")

    def wait_for_move_in():
        """ """
        qtbot.mouseClick(conf.pushButton_move_step_in, Qt.MouseButton.LeftButton)
        assert len(prot.loop_steps) == 1

    qtbot.waitUntil(wait_for_move_in)
    with qtbot.waitSignal(conf.accepted) as blocker:
        conf.accept()
    prot.name = "test_for_loop_protocol"
    assert "For Loop (For_Loop)" in prot.loop_step_dict
    assert prot.loop_steps[0].has_children
    assert prot.loop_steps[0].children[0].full_name == "Wait (Wait)"
    catalog_maker(tmp_path)
    run_test_protocol(tmp_path, prot)


def test_gradient_descent(qtbot, tmp_path):
    """Opens the config for "Gradient Descent" tries to configure it for the
    demo instrument and tries to run a protocol with this step."""
    ensure_demo_in_devices()
    from nomad_camels.loop_steps import gradient_descent

    conf = protocol_config.Protocol_Config()
    qtbot.addWidget(conf)
    action = get_action_from_name(conf.add_actions, "Gradient Descent")
    action.trigger()
    conf_widge = conf.loop_step_configuration_widget
    assert isinstance(conf_widge, gradient_descent.Gradient_Descent_Config)
    conf_widge.sub_widget.lineEdit_momentum.setText("1")
    conf_widge.sub_widget.lineEdit_threshold.setText("1e-3")
    conf_widge.sub_widget.lineEdit_max_val.setText("1")
    conf_widge.sub_widget.lineEdit_min_val.setText("1")
    conf_widge.sub_widget.lineEdit_opt_func.setText("demo_instrument_detectorY")
    conf_widge.sub_widget.lineEdit_max_n_steps.setText("25")
    conf_widge.sub_widget.lineEdit_largest_step.setText("1")
    conf_widge.sub_widget.lineEdit_smallest_step.setText("0.01")
    conf_widge.sub_widget.lineEdit_learning_rate.setText("1")
    conf_widge.sub_widget.lineEdit_starting_val.setText("0")
    conf_widge.sub_widget.comboBox_extremum_type.setCurrentText("Maximum")
    conf_widge.sub_widget.comboBox_output_channel.setCurrentText(
        "demo_instrument_motorY"
    )
    with qtbot.waitSignal(conf.accepted) as blocker:
        conf.accept()
    prot = conf.protocol
    prot.loop_steps[0].read_channels.append("demo_instrument_detectorY")
    prot.name = "test_gradient_descent_protocol"
    assert "Gradient Descent (Gradient_Descent)" in prot.loop_step_dict
    catalog_maker(tmp_path)
    run_test_protocol(tmp_path, prot)


def test_if_and_set_variables(qtbot, tmp_path):
    """Opens the config for "If" tries to configure to run the correct way if
    the variables have the correct value, which is done by a "Set Variables" step."""
    from nomad_camels.loop_steps import set_variables

    conf = protocol_config.Protocol_Config()
    prot = conf.protocol
    qtbot.addWidget(conf)
    qtbot.mouseClick(
        conf.general_settings.pushButton_add_variable, Qt.MouseButton.LeftButton
    )
    conf.general_settings.variable_table.model.item(0, 0).setText("condition")
    conf.general_settings.variable_table.model.item(0, 1).setText("1")
    single_variable_if(qtbot, conf, 1)
    assert "If (If)" in prot.loop_step_dict
    assert "If_Sub (If_condition != 1)" in prot.loop_step_dict
    assert "Elif_Sub (If_condition == 1)" in prot.loop_step_dict
    assert "Else_Sub (If_Else)" in prot.loop_step_dict

    action = get_action_from_name(conf.add_actions, "Set Variables")
    action.trigger()

    def wait_selection():
        """ """
        select_step_by_name(conf, f"Set Variables (Set_Variables)")
        conf.tree_click_sequence()
        assert isinstance(
            conf.loop_step_configuration_widget, set_variables.Set_Variables_Config
        )

    qtbot.waitUntil(wait_selection)
    conf_widge = conf.loop_step_configuration_widget
    assert isinstance(
        conf.loop_step_configuration_widget, set_variables.Set_Variables_Config
    )
    conf_widge.variables_table.add(["condition", "0"])

    single_variable_if(qtbot, conf, 0, 2, 1, 2)
    with qtbot.waitSignal(conf.accepted) as blocker:
        conf.accept()
    catalog_maker(tmp_path)
    run_test_protocol(tmp_path, prot)


def test_nd_sweep():
    """ """
    pass


def test_read_channels(qtbot, tmp_path):
    """Opens the config for "Read Channels" tries to configure it for the
    demo instrument and tries to run a protocol with this step."""
    ensure_demo_in_devices()
    from nomad_camels.loop_steps import read_channels

    conf = protocol_config.Protocol_Config()
    qtbot.addWidget(conf)
    action = get_action_from_name(conf.add_actions, "Read Channels")
    action.trigger()
    conf_widge = conf.loop_step_configuration_widget
    assert isinstance(conf_widge, read_channels.Read_Channels_Config)
    table = conf_widge.sub_widget.read_table.tableWidget_channels
    row = get_row_from_channel_table("demo_instrument_detectorX", table)
    table.item(row, 0).setCheckState(Qt.CheckState.Checked)
    with qtbot.waitSignal(conf.accepted) as blocker:
        conf.accept()
    prot = conf.protocol
    prot.name = "test_read_channels_protocol"
    assert "Read Channels (Read_Channels)" in prot.loop_step_dict
    assert prot.loop_steps[0].channel_list == ["demo_instrument_detectorX"]
    catalog_maker(tmp_path)
    run_test_protocol(tmp_path, prot)


def test_run_subprotocol():
    """ """
    pass


def test_set_channels(qtbot, tmp_path):
    """Opens the config for "Set Channels" tries to configure it for the
    demo instrument and tries to run a protocol with this step."""
    ensure_demo_in_devices()
    from nomad_camels.loop_steps import set_channels

    conf = protocol_config.Protocol_Config()
    qtbot.addWidget(conf)
    action = get_action_from_name(conf.add_actions, "Set Channels")
    action.trigger()
    conf_widge = conf.loop_step_configuration_widget
    assert isinstance(conf_widge, set_channels.Set_Channels_Config)
    table = conf_widge.sub_widget.tableWidget_channels
    row = get_row_from_channel_table("demo_instrument_motorX", table)
    table.item(row, 0).setCheckState(Qt.CheckState.Checked)
    table.item(row, 2).setText("1")
    with qtbot.waitSignal(conf.accepted) as blocker:
        conf.accept()
    prot = conf.protocol
    prot.name = "test_set_channels_protocol"
    assert "Set Channels (Set_Channels)" in prot.loop_step_dict
    assert prot.loop_steps[0].channels_values == {
        "Channels": ["demo_instrument_motorX"],
        "Values": ["1"],
    }
    catalog_maker(tmp_path)
    run_test_protocol(tmp_path, prot)


def test_simple_sweep_with_plot_and_fit(qtbot, tmp_path):
    """Opens the config for "Simple Sweep" tries to configure it for the
    demo instrument. Further it adds a plot and a fit to the sweep and tries to
    run a protocol with this step."""
    ensure_demo_in_devices()
    from nomad_camels.loop_steps import simple_sweep
    from nomad_camels.frontpanels import plot_definer

    conf = protocol_config.Protocol_Config()
    qtbot.addWidget(conf)
    action = get_action_from_name(conf.add_actions, "Simple Sweep")
    action.trigger()
    conf_widge = conf.loop_step_configuration_widget
    assert isinstance(conf_widge, simple_sweep.Simple_Sweep_Config)
    conf_widge.sweep_widget.lineEdit_start.setText("-10")
    conf_widge.sweep_widget.lineEdit_stop.setText("10")
    conf_widge.sweep_widget.lineEdit_n_points.setText("21")
    conf_widge.comboBox_sweep_channel.setCurrentText("demo_instrument_motorY")

    table = conf_widge.read_table.tableWidget_channels
    row = get_row_from_channel_table("demo_instrument_detectorY", table)
    table.item(row, 0).setCheckState(Qt.CheckState.Checked)
    row = get_row_from_channel_table("demo_instrument_motorY", table)
    table.item(row, 0).setCheckState(Qt.CheckState.Checked)

    fit = plot_definer.Fit_Info(
        True, "Gaussian", x="demo_instrument_motorY", y="demo_instrument_detectorY"
    )
    plot = plot_definer.Plot_Info(
        x_axis="demo_instrument_motorY",
        y_axes={"formula": ["demo_instrument_detectorY"], "axis": ["left"]},
        fits=[fit],
    )
    conf_widge.plot_widge.plot_data = [plot]

    with qtbot.waitSignal(conf.accepted) as blocker:
        conf.accept()
    prot = conf.protocol
    prot.name = "test_simple_sweep_protocol"
    assert "Simple Sweep (Simple_Sweep)" in prot.loop_step_dict
    catalog_maker(tmp_path)
    run_test_protocol(tmp_path, prot)


def test_trigger_and_read_channels(qtbot, tmp_path):
    """Opens the config for "Read Channels" tries to configure it with a split
    triggering of the channels, then configures the trigger-step and tries to
    run a protocol with these steps."""
    ensure_demo_in_devices()
    from nomad_camels.loop_steps import read_channels

    conf = protocol_config.Protocol_Config()
    qtbot.addWidget(conf)
    prot = conf.protocol
    variables_handling.current_protocol = prot
    action = get_action_from_name(conf.add_actions, "Read Channels")
    action.trigger()
    conf_widge = conf.loop_step_configuration_widget
    assert isinstance(conf_widge, read_channels.Read_Channels_Config)
    conf_widge.sub_widget.checkBox_split_trigger.setChecked(True)
    conf_widge.sub_widget.checkBox_read_all.setChecked(True)

    action = get_action_from_name(conf.add_actions, "Trigger Channels")
    action.trigger()
    select_step_by_name(conf, "Trigger Channels (Trigger_Channels)")
    conf.tree_click_sequence()
    conf_widge = conf.loop_step_configuration_widget
    assert isinstance(conf_widge, read_channels.Trigger_Channels_Config)

    def wait_for_move():
        """ """
        qtbot.mouseClick(conf.pushButton_move_step_up, Qt.MouseButton.LeftButton)
        assert isinstance(prot.loop_steps[0], read_channels.Trigger_Channels_Step)

    qtbot.waitUntil(wait_for_move)

    with qtbot.waitSignal(conf.accepted) as blocker:
        conf.accept()
    prot.name = "test_trigger_protocol"
    assert "Read Channels (Read_Channels)" in prot.loop_step_dict
    assert prot.loop_steps[1].read_all
    assert "Trigger Channels (Trigger_Channels)" in prot.loop_step_dict
    assert prot.loop_steps[0].read_step == "Read Channels (Read_Channels)"

    catalog_maker(tmp_path)
    run_test_protocol(tmp_path, prot)


def test_while_loop(qtbot, tmp_path):
    """Opens the config for a "While Loop" and tries to configure it with the
    condition being `While_Loop_Count < 5`. Into the loop, a wait-step is added.
    Tries to run a protocol with this step."""
    from nomad_camels.loop_steps import for_while_loops

    conf = protocol_config.Protocol_Config()
    qtbot.addWidget(conf)
    action = get_action_from_name(conf.add_actions, "While Loop")
    action.trigger()
    conf_widge = conf.loop_step_configuration_widget
    assert isinstance(conf_widge, for_while_loops.While_Loop_Step_Config)
    conf_widge.sub_widget.lineEdit_condition.setText("While_Loop_Count < 5")

    action = get_action_from_name(conf.add_actions, "Wait")
    action.trigger()
    prot = conf.protocol

    select_step_by_name(conf, "Wait (Wait)")

    def wait_for_move_in():
        """ """
        qtbot.mouseClick(conf.pushButton_move_step_in, Qt.MouseButton.LeftButton)
        assert len(prot.loop_steps) == 1

    qtbot.waitUntil(wait_for_move_in)
    with qtbot.waitSignal(conf.accepted) as blocker:
        conf.accept()
    prot.name = "test_while_loop_protocol"
    assert "While Loop (While_Loop)" in prot.loop_step_dict
    assert prot.loop_steps[0].has_children
    assert prot.loop_steps[0].children[0].full_name == "Wait (Wait)"
    catalog_maker(tmp_path)
    run_test_protocol(tmp_path, prot)


def test_wait(qtbot, tmp_path):
    """Opens the config for "Wait" tries to configure it to a wait time of
    1.0 second and tries to run a protocol with this step."""
    from nomad_camels.loop_steps import wait_loop_step

    conf = protocol_config.Protocol_Config()
    qtbot.addWidget(conf)
    action = get_action_from_name(conf.add_actions, "Wait")
    action.trigger()
    conf_widge = conf.loop_step_configuration_widget
    assert isinstance(conf_widge, wait_loop_step.Wait_Loop_Step_Config)
    conf_widge.sub_widget.lineEdit_duration.setText("1.0")
    with qtbot.waitSignal(conf.accepted) as blocker:
        conf.accept()
    prot = conf.protocol
    prot.name = "test_wait_protocol"
    assert "Wait (Wait)" in prot.loop_step_dict
    assert prot.loop_steps[0].wait_time == "1.0"
    catalog_maker(tmp_path)
    run_test_protocol(tmp_path, prot)


def single_variable_if(qtbot, conf, wait_in=1, n_prompt=0, n_if=0, len_prot=0):
    """

    Parameters
    ----------
    qtbot :

    conf :

    wait_in :
         (Default value = 1)
    n_prompt :
         (Default value = 0)
    n_if :
         (Default value = 0)
    len_prot :
         (Default value = 0)

    Returns
    -------

    """
    from nomad_camels.loop_steps import if_step

    action = get_action_from_name(conf.add_actions, "If")
    action.trigger()

    def wait_selection():
        """ """
        if n_if:
            select_step_by_name(conf, f"If (If_{n_if})")
        else:
            select_step_by_name(conf, f"If (If)")
        conf.tree_click_sequence()
        assert isinstance(conf.loop_step_configuration_widget, if_step.If_Step_Config)

    qtbot.waitUntil(wait_selection)
    conf_widge = conf.loop_step_configuration_widget
    conf_widge.sub_widget.lineEdit_condition.setText("condition != 1")
    conf_widge.sub_widget.elif_table.add("condition == 1")
    conf_widge.sub_widget.checkBox_use_else.setChecked(True)
    nam = "Prompt" if 0 != wait_in else "Wait"
    action = get_action_from_name(conf.add_actions, nam)
    action.trigger()
    prot = conf.protocol

    if n_prompt > 0 and nam == "Prompt":
        select_step_by_name(conf, f"{nam} ({nam}_{n_prompt})")
        n_prompt += 1
    elif nam == "Wait" and n_if:
        select_step_by_name(conf, f"{nam} ({nam}_{n_if})")
    else:
        select_step_by_name(conf, f"{nam} ({nam})")

    def wait_for_move():
        """ """
        qtbot.mouseClick(conf.pushButton_move_step_in, Qt.MouseButton.LeftButton)
        qtbot.mouseClick(conf.pushButton_move_step_out, Qt.MouseButton.LeftButton)
        qtbot.mouseClick(conf.pushButton_move_step_up, Qt.MouseButton.LeftButton)
        qtbot.mouseClick(conf.pushButton_move_step_up, Qt.MouseButton.LeftButton)
        qtbot.mouseClick(conf.pushButton_move_step_in, Qt.MouseButton.LeftButton)
        assert len(prot.loop_steps) == 1 + len_prot

    qtbot.waitUntil(wait_for_move)

    nam = "Prompt" if 1 != wait_in else "Wait"
    action = get_action_from_name(conf.add_actions, nam)
    action.trigger()
    if n_prompt > 0 and nam == "Prompt":
        select_step_by_name(conf, f"{nam} ({nam}_{n_prompt})")
        n_prompt += 1
    elif nam == "Wait" and n_if:
        select_step_by_name(conf, f"{nam} ({nam}_{n_if})")
    else:
        select_step_by_name(conf, f"{nam} ({nam})")

    def wait_for_move():
        """ """
        qtbot.mouseClick(conf.pushButton_move_step_in, Qt.MouseButton.LeftButton)
        qtbot.mouseClick(conf.pushButton_move_step_out, Qt.MouseButton.LeftButton)
        qtbot.mouseClick(conf.pushButton_move_step_up, Qt.MouseButton.LeftButton)
        qtbot.mouseClick(conf.pushButton_move_step_in, Qt.MouseButton.LeftButton)
        assert len(prot.loop_steps) == 1 + len_prot

    qtbot.waitUntil(wait_for_move)

    action = get_action_from_name(conf.add_actions, "Prompt")
    action.trigger()
    if n_prompt == 0:
        n_prompt += 1
    select_step_by_name(conf, f"Prompt (Prompt_{n_prompt})")

    def wait_for_move():
        """ """
        qtbot.mouseClick(conf.pushButton_move_step_in, Qt.MouseButton.LeftButton)
        assert len(prot.loop_steps) == 1 + len_prot

    qtbot.waitUntil(wait_for_move)


def select_step_by_name(conf, name):
    """

    Parameters
    ----------
    conf :

    name :


    Returns
    -------

    """
    model = conf.treeView_protocol_sequence.model()
    conf.treeView_protocol_sequence.selectionModel().clearSelection()
    index = getItemIndex(model, name)
    conf.treeView_protocol_sequence.selectionModel().select(
        index, QItemSelectionModel.Select
    )
    conf.check_movability()


def get_row_from_channel_table(name, table):
    """

    Parameters
    ----------
    name :

    table :


    Returns
    -------

    """
    for i in range(table.rowCount()):
        row_name = table.item(i, 1).text()
        if name == row_name:
            return i


def ensure_demo_in_devices():
    """ """
    instr, packs = instrument_installer.getInstalledDevices(True, True)
    if "demo_instrument" not in instr:
        instrument_installer.install_instrument("demo_instrument")
        instr, packs = instrument_installer.getInstalledDevices(True, True)
        assert "demo_instrument" in instr
    if "demo_instrument" not in variables_handling.devices:
        inst = packs["demo_instrument"].subclass()
        variables_handling.devices["demo_instrument"] = inst
    assert "demo_instrument" in variables_handling.devices
    variables_handling.channels.clear()
    for key, dev in variables_handling.devices.items():
        variables_handling.channels.update(dev.get_channels())


def catalog_maker(tmp_path):
    """

    Parameters
    ----------
    tmp_path :


    Returns
    -------

    """
    if tmp_path is None:
        tmp_path = ""
    if "test_catalog" not in list(databroker.catalog):
        make_catalog.make_yml(tmp_path, "test_catalog")


def run_test_protocol(tmp_path, protocol):
    """

    Parameters
    ----------
    tmp_path :

    protocol :


    Returns
    -------

    """
    from nomad_camels.bluesky_handling import protocol_builder

    file = tmp_path / (protocol.name + ".py")
    savepath = tmp_path / (protocol.name + ".nxs")
    protocol_builder.build_protocol(protocol, file, savepath, "test_catalog")
    sys.path.append(str(tmp_path))
    py_package = importlib.import_module(protocol.name)
    py_package.main()
    assert os.path.isfile(savepath)


def get_action_from_name(actions, name):
    """

    Parameters
    ----------
    actions :

    name :


    Returns
    -------

    """
    for act in actions:
        if act.text() == name:
            return act
    return None



import importlib.metadata
import os.path
import json
import sys

import databroker
import h5py
from datetime import datetime as dt

import numpy as np
import xarray

from PySide6.QtWidgets import (
    QDialog,
    QComboBox,
    QPushButton,
    QGridLayout,
    QLabel,
    QCheckBox,
)


def recourse_entry_dict(entry, metadata):
    """Recoursively makes the metadata to a dictionary.

    Parameters
    ----------
    entry :

    metadata :


    Returns
    -------

    """
    # TODO check if actually necessary
    for key, val in metadata.items():
        if isinstance(val, databroker.core.Start) or isinstance(
            val, databroker.core.Stop
        ):
            val = dict(val)
            stamp = val["time"]
            val["time"] = timestamp_to_ISO8601(stamp)
            # stamp = rundict['metadata_stop']['time']
            # rundict['metadata_stop']['time'] = timestamp_to_ISO8601(stamp)
        if type(val) is dict:
            if key == "start":
                sub_entry = entry
            else:
                sub_entry = entry.create_group(key)
            recourse_entry_dict(sub_entry, val)
        elif type(val) is list:
            no_dict = False
            for i, value in enumerate(val):
                if isinstance(value, dict):
                    sub_entry = entry.create_group(f"{key}_{i}")
                    recourse_entry_dict(sub_entry, value)
                # else:
                #     # entry.attrs[f'{key}_{i}'] = val
                else:
                    no_dict = True
                    break
            if no_dict:
                if any(isinstance(item, str) for item in val):
                    entry[key] = np.array(val).astype("S")
                else:
                    entry[key] = val
        elif val is None:
            continue
        else:
            # entry.attrs[key] = val
            entry[key] = val


def broker_to_hdf5(runs, filename, additional_data=None):
    """Puts the given `runs` into `filename`, containing the run's
    metadata and the dataset.

    Parameters
    ----------
    runs :

    filename :

    additional_data :
         (Default value = None)

    Returns
    -------

    """
    if not os.path.isdir(os.path.dirname(filename)):
        os.makedirs(os.path.dirname(filename))
    if not isinstance(runs, list):
        runs = [runs]
    for run in runs:
        metadata = run.metadata
        start_time = timestamp_to_ISO8601(metadata["start"]["time"])
        with h5py.File(filename, "a") as file:
            entry = file.create_group(start_time)
            recourse_entry_dict(entry, metadata)
            additional_data = additional_data or {}
            recourse_entry_dict(entry, additional_data)
            for stream in run:
                dataset = run[stream].read()
                # entry[stream] = dataset.to_pandas()
                # return
                group = entry.create_group(stream)
                for coord in dataset.coords:
                    if coord == "time":
                        isos = []
                        for t in dataset[coord].values:
                            isos.append(timestamp_to_ISO8601(t))
                        since = dataset[coord].to_numpy()
                        since -= metadata["start"]["time"]
                        group["time_since_start"] = since
                        group["time"] = isos
                    else:
                        group[coord] = dataset[coord]
                stream_meta = run[stream].metadata
                for col in dataset:
                    group[col] = dataset[col]
                if "descriptors" in stream_meta:
                    dat = stream_meta["descriptors"][0]["data_keys"]
                    for key, val in dat.items():
                        if isinstance(val, dict):
                            for k, v in val.items():
                                group[key].attrs[k] = v


def export_run(
    filename,
    run_number=-1,
    plot_data=None,
    additional_data=None,
    session_name="",
    export_to_csv=False,
    export_to_json=False,
    catalog_name="CAMELS_CATALOG",
    new_file_each_run=False,
):
    """TODO"""
    catalog = databroker.catalog[catalog_name]
    run = catalog[run_number]
    from nomad_camels.bluesky_handling.helper_functions import export_function

    export_function(
        [run],
        filename,
        True,
        new_file_each_run,
        export_to_csv,
        export_to_json,
        plot_data,
    )
    # broker_to_NX(
    #     [run],
    #     filename,
    #     plot_data,
    #     additional_data,
    #     session_name,
    #     export_to_csv,
    #     export_to_json,
    #     new_file_each_run,
    # )


def broker_to_NX(
    runs,
    filename,
    plot_data=None,
    additional_data=None,
    session_name="",
    export_to_csv=False,
    export_to_json=False,
    new_file_each_run=False,
):
    """

    Parameters
    ----------
    runs :

    filename :

    plot_data :
         (Default value = None)
    additional_data :
         (Default value = None)
    session_name :
         (Default value = '')
    export_to_csv :
         (Default value = False)
    export_to_json :
         (Default value = False)

    Returns
    -------

    """
    if not os.path.isdir(os.path.dirname(filename)):
        os.makedirs(os.path.dirname(filename))
    if not isinstance(runs, list):
        runs = [runs]
    for run in runs:
        metadata = run.metadata
        meta_start = dict(metadata["start"])
        meta_stop = {"time": None}
        if "stop" in metadata and metadata["stop"]:
            meta_stop = dict(metadata["stop"])
        st_time = meta_start.pop("time")
        start_time = timestamp_to_ISO8601(st_time)
        end_time = timestamp_to_ISO8601(meta_stop.pop("time"))
        entry_name = f"{session_name}_{start_time}" if session_name else start_time
        entry_name_non_iso = clean_filename(entry_name)
        # check if the filename already exists, if yes, add entry_name to filename
        if new_file_each_run and os.path.isfile(filename):
            filename = os.path.splitext(filename)[0] + f"_{entry_name_non_iso}.nxs"
        filename = os.path.join(
            os.path.abspath(os.path.dirname(filename)),
            os.path.normpath(
                f"{clean_filename(os.path.splitext(os.path.basename(filename))[0])}.nxs"
            ),
        )
        if export_to_json:
            if not os.path.isdir(filename.split(".")[0]):
                os.makedirs(filename.split(".")[0])
            with open(
                f'{filename.split(".")[0]}/{entry_name_non_iso}_metadata.json',
                "w",
                encoding="utf-8",
            ) as json_file:
                json.dump(meta_start, json_file, indent=2)
        with h5py.File(filename, "a") as file:
            entry = file.create_group(entry_name)
            entry.attrs["NX_class"] = "NXentry"
            entry["definition"] = "NXsensor_scan"
            entry["start_time"] = start_time
            entry["end_time"] = end_time
            if "description" in meta_start:
                desc = meta_start.pop("description")
                entry["experiment_description"] = desc
            if "identifier" in meta_start:
                ident = meta_start.pop("identifier")
                entry["experiment_identifier"] = ident
            proc = entry.create_group("process")
            proc.attrs["NX_class"] = "NXprocess"
            proc["program"] = "NOMAD CAMELS"
            proc["program"].attrs["version"] = "0.1"
            proc["program"].attrs[
                "program_url"
            ] = "https://github.com/FAU-LAP/NOMAD-CAMELS"
            version_dict = meta_start.pop("versions")
            vers_group = proc.create_group("versions")
            py_environment = proc.create_group("python_environment")
            py_environment.attrs["python_version"] = sys.version
            for x in importlib.metadata.distributions():
                name = x.metadata["Name"]
                if name not in py_environment.keys():
                    py_environment[x.metadata["Name"]] = x.version
                # except Exception as e:
                #     print(e, x.metadata['Name'])
            recourse_entry_dict(vers_group, version_dict)
            user = entry.create_group("user")
            user.attrs["NX_class"] = "NXuser"
            recourse_entry_dict(user, meta_start.pop("user"))
            sample = entry.create_group("sample")
            sample.attrs["NX_class"] = "NXsample"
            recourse_entry_dict(sample, meta_start.pop("sample"))

            instr = entry.create_group("instrument")
            instr.attrs["NX_class"] = "NXinstrument"
            for dev, dat in meta_start.pop("devices").items():
                dev_group = instr.create_group(dev)
                dev_group.attrs["NX_class"] = "NXsensor"
                if "idn" in dat:
                    dev_group["model"] = dat.pop("idn")
                else:
                    dev_group["model"] = dat["device_class_name"]
                dev_group["name"] = dat.pop("device_class_name")
                dev_group["short_name"] = dev
                settings = dev_group.create_group("settings")
                recourse_entry_dict(settings, dat)

            recourse_entry_dict(entry, meta_start)

            data_entry = entry.create_group("data")
            data_entry.attrs["NX_class"] = "NXdata"
            for stream in run:
                if "_fits_readying_" in stream:
                    continue
                dataset = run[stream].read()
                if export_to_csv:
                    if not os.path.isdir(
                        f'{filename.split(".")[0]}/{entry_name_non_iso}'
                    ):
                        os.makedirs(f'{filename.split(".")[0]}/{entry_name_non_iso}')
                    try:
                        df = dataset.to_dataframe()
                        df.to_csv(
                            f'{filename.split(".")[0]}/{entry_name_non_iso}/{stream}.csv',
                            sep=";",
                        )
                    except Exception as e:
                        print(e)
                if stream == "primary":
                    group = data_entry
                else:
                    group = data_entry.create_group(stream)
                    group.attrs["NX_class"] = "NXdata"
                for coord in dataset.coords:
                    if coord == "time":
                        isos = []
                        for t in dataset[coord].values:
                            isos.append(timestamp_to_ISO8601(t))
                        since = dataset[coord].to_numpy()
                        since -= st_time
                        group["time_since_start"] = since
                        group["time"] = isos
                    else:
                        group[coord] = dataset[coord]
                stream_meta = run[stream].metadata
                for col in dataset:
                    # Enables CAMELS to read strings from channels
                    if str(dataset[col].dtype).startswith("<U"):
                        group[col] = dataset[col].astype(bytes)
                    else:
                        group[col] = dataset[col]
                if "descriptors" in stream_meta:
                    dat = stream_meta["descriptors"][0]["data_keys"]
                    for key, val in dat.items():
                        if isinstance(val, dict):
                            for k, v in val.items():
                                try:
                                    group[key].attrs[k] = v
                                except Exception as e:
                                    print(
                                        f"could not add value {v} to metadata with name {k}\n{e}"
                                    )
                if not plot_data:
                    continue
                axes = []
                signals = []
                for plot in plot_data:
                    if plot.stream_name == stream and hasattr(plot, "x_name"):
                        if plot.x_name not in axes:
                            axes.append(plot.x_name)
                        if hasattr(plot, "z_name"):
                            if plot.y_name not in axes:
                                axes.append(plot.y_name)
                            if plot.z_name not in signals:
                                signals.append(plot.z_name)
                        else:
                            for y in plot.y_names:
                                if y not in signals:
                                    signals.append(y)
                        if not hasattr(plot, "liveFits") or not plot.liveFits:
                            continue
                        fit_group = group.require_group("fits")
                        for fit in plot.liveFits:
                            if not fit.results:
                                continue
                            fg = fit_group.require_group(fit.name)
                            param_names = []
                            param_values = []
                            covars = []
                            timestamps = []
                            for t, res in fit.results.items():
                                timestamps.append(float(t))
                                if res.covar is None:
                                    covar = np.ones(
                                        (len(res.best_values), len(res.best_values))
                                    )
                                    covar *= np.nan
                                else:
                                    covar = res.covar
                                covars.append(covar)
                                if not param_names:
                                    param_names = res.model.param_names
                                param_values.append(res.params)
                            fg.attrs["param_names"] = param_names
                            timestamps, covars, param_values = sort_by_list(
                                timestamps, [covars, param_values]
                            )
                            isos = []
                            for t in timestamps:
                                isos.append(timestamp_to_ISO8601(t))
                            fg["time"] = isos
                            since = np.array(timestamps)
                            since -= st_time
                            fg["time_since_start"] = since
                            fg["covariance"] = covars
                            fg["covariance"].attrs["parameters"] = param_names[
                                : len(covars[0])
                            ]
                            param_values = get_param_dict(param_values)
                            for p, v in param_values.items():
                                fg[p] = v
                            for name, val in fit.additional_data.items():
                                fg[name] = val
                if signals:
                    group.attrs["signal"] = signals[0]
                    if len(signals) > 1:
                        group.attrs["auxiliary_signals"] = signals[1:]
                group.attrs["axes"] = axes
                group.attrs["NX_class"] = "NXdata"
    # print('Successfully saved protocol data!')
    return filename


def export_h5_to_csv_json(
    filename, entry_name=None, export_data=True, export_metadata=True, export_path=None
):
    # read the h5 file and list the entries

    with h5py.File(filename, "r") as file:
        entries = list(file.keys())
    if entry_name is not None and entry_name in entries:
        entries = [entry_name]
    if export_path is None:
        fpath = os.path.join(
            os.path.dirname(filename), os.path.basename(filename).split(".")[0]
        )
    else:
        fpath = export_path
    for entry_name in entries:
        with h5py.File(filename, "r") as file:
            entry = file[entry_name]
            entry_name_non_iso = clean_filename(entry_name)
            if export_metadata:
                metadata = h5_group_to_dict(entry)
                fname = os.path.join(fpath, f"{entry_name_non_iso}_metadata.json")
                if not os.path.isdir(os.path.dirname(fname)):
                    os.makedirs(os.path.dirname(fname))
                with open(fname, "w", encoding="utf-8") as json_file:
                    json.dump(metadata, json_file, cls=NumpyEncoder, indent=2)
            if export_data:
                if "data" not in entry:
                    continue
                data = entry["data"]
                export_h5_group_to_csv(
                    data, os.path.join(fpath, entry_name_non_iso, "primary.csv")
                )


class NumpyEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        elif isinstance(obj, np.bool_):
            return bool(obj)
        elif isinstance(obj, bytes):
            import base64

            return base64.b64encode(obj).decode("utf-8")
        else:
            return super(NumpyEncoder, self).default(obj)


def h5_group_to_dict(group):
    """

    Parameters
    ----------
    group :


    Returns
    -------

    """
    data = {}
    for k in group.keys():
        if k == "data":
            continue
        if isinstance(group[k], h5py.Group):
            data[k] = h5_group_to_dict(group[k])
        else:
            # read dataset and handle scalar data
            if len(group[k].shape) == 0:
                data[k] = group[k][()]
            else:
                data[k] = group[k][:].tolist()
            # if data[k] is of type bytes, convert it to string
            if isinstance(data[k], bytes):
                data[k] = data[k].decode("utf-8")
    return data


def export_h5_group_to_csv(group, filename):
    import pandas as pd

    arrs = {}
    if not os.path.isdir(os.path.dirname(filename)):
        os.makedirs(os.path.dirname(filename))
    for k in group.keys():
        if isinstance(group[k], h5py.Group):
            export_h5_group_to_csv(group[k], f'{filename.split(".")[0]}_{k}.csv')
        elif isinstance(group[k], h5py.Dataset):
            # convert the dataset to a numpy array
            arr = group[k][:]
            # if the array is 2D save it as a dataframe
            if len(arr.shape) == 2:
                df = pd.DataFrame(arr)
                df.to_csv(f'{filename.split(".")[0]}_{k}.csv', sep=",")
            elif len(arr.shape) > 2:
                # Create a MultiIndex for an unknown number of dimensions
                index = pd.MultiIndex.from_product(
                    [range(i) for i in arr.shape],
                    names=[f"dim{i}" for i in range(len(arr.shape))],
                )
                # Flatten the array and create a DataFrame
                arr_flat = arr.flatten()
                df = pd.DataFrame(arr_flat, index=index)
                df.to_csv(f'{filename.split(".")[0]}_{k}.csv', sep=",")
            else:
                arrs[k] = arr
    if arrs:
        try:
            df = pd.DataFrame(arrs)
        except Exception as e:
            print(e)
            return
        df.to_csv(filename, sep=",")


class ExportH5_dialog(QDialog):
    def __init__(self, parent=None):
        from nomad_camels.ui_widgets.path_button_edit import Path_Button_Edit

        super().__init__(parent)
        self.setWindowTitle("Export H5 - NOMAD CAMELS")
        layout = QGridLayout()
        self.setLayout(layout)

        self.filename = Path_Button_Edit()
        self.filename.path_changed.connect(self.update_entries)

        label_entry = QLabel("Entry")
        self.entry_name = QComboBox()
        self.export_button = QPushButton("Export")
        self.export_button.clicked.connect(self.accept)

        self.checkbox_all_entries = QCheckBox("export all entries")
        self.checkbox_all_entries.stateChanged.connect(self.all_entries)

        self.checkbox_export_to_dir = QCheckBox("export to directory of hdf5 file")
        self.export_path = Path_Button_Edit(select_directory=True)
        self.export_path.setEnabled(False)
        self.checkbox_export_to_dir.stateChanged.connect(self.export_path.setDisabled)
        self.checkbox_export_to_dir.setChecked(True)

        self.checkbox_data = QCheckBox("export data")
        self.checkbox_data.setChecked(True)
        self.checkbox_metadata = QCheckBox("export metadata")
        self.checkbox_metadata.setChecked(True)

        self.cancel_button = QPushButton("Cancel")
        self.cancel_button.clicked.connect(self.reject)

        layout.addWidget(QLabel("hdf5 file"), 0, 0)
        layout.addWidget(self.filename, 0, 1)
        layout.addWidget(label_entry, 1, 0)
        layout.addWidget(self.entry_name, 1, 1)
        layout.addWidget(self.checkbox_all_entries, 2, 0, 1, 2)
        layout.addWidget(self.checkbox_data, 3, 0)
        layout.addWidget(self.checkbox_metadata, 3, 1)
        layout.addWidget(self.checkbox_export_to_dir, 5, 0)
        layout.addWidget(self.export_path, 5, 1)
        layout.addWidget(self.export_button, 10, 0)
        layout.addWidget(self.cancel_button, 10, 1)

        self.show()
        self.adjustSize()

    def all_entries(self):
        if self.checkbox_all_entries.isChecked():
            self.entry_name.setEnabled(False)
        else:
            self.entry_name.setEnabled(True)

    def update_entries(self):
        self.entry_name.clear()
        if not self.filename.get_path():
            return
        with h5py.File(self.filename.get_path(), "r") as file:
            entries = list(file.keys())
        self.entry_name.addItems(entries)

    def accept(self):
        if not self.filename.get_path():
            return
        if self.checkbox_all_entries.isChecked():
            entry = None
        else:
            entry = self.entry_name.currentText()
        export_data = self.checkbox_data.isChecked()
        export_metadata = self.checkbox_metadata.isChecked()
        export_path = (
            self.export_path.get_path()
            if not self.checkbox_export_to_dir.isChecked()
            else None
        )
        export_h5_to_csv_json(
            self.filename.get_path(),
            entry_name=entry,
            export_data=export_data,
            export_metadata=export_metadata,
            export_path=export_path,
        )
        from nomad_camels.ui_widgets.warn_popup import WarnPopup

        WarnPopup(
            text="Successfully exported data to csv and json",
            title="Export successful",
            info_icon=True,
        )
        super().accept()


def get_param_dict(param_values):
    """

    Parameters
    ----------
    param_values :


    Returns
    -------

    """
    p_s = {}
    for vals in param_values:
        for k in vals:
            if k in p_s:
                p_s[k].append(vals[k].value)
            else:
                p_s[k] = [vals[k].value]
    return p_s


def sort_by_list(sort_list, other_lists):
    """

    Parameters
    ----------
    sort_list :

    other_lists :


    Returns
    -------

    """
    s_list = sorted(zip(sort_list, *other_lists), key=lambda x: x[0])
    return zip(*s_list)


def broker_to_dict(runs, to_iso_time=True):
    """Puts the runs into a dictionary.

    Parameters
    ----------
    runs :

    to_iso_time :
         (Default value = True)

    Returns
    -------

    """
    dicts = []
    if not isinstance(runs, list):
        runs = [runs]
    for run in runs:
        data = {}
        for stream in run:
            data = run[stream].read()
            if isinstance(data, xarray.Dataset):
                data = data.to_array()
        rundict = {
            "metadata_start": dict(run.metadata["start"]),
            "metadata_stop": dict(run.metadata["stop"]),
            "data": data,
        }
        if to_iso_time:
            stamp = rundict["metadata_start"]["time"]
            rundict["metadata_start"]["time"] = timestamp_to_ISO8601(stamp)
            stamp = rundict["metadata_stop"]["time"]
            rundict["metadata_stop"]["time"] = timestamp_to_ISO8601(stamp)
        dicts.append(rundict)
    return dicts


def timestamp_to_ISO8601(timestamp):
    """

    Parameters
    ----------
    timestamp :


    Returns
    -------

    """
    if timestamp is None:
        return "None"
    from_stamp = dt.fromtimestamp(timestamp)
    return from_stamp.astimezone().isoformat()


def clean_filename(filename):
    """
    cleans the filename from characters that are not allowed

    Parameters
    ----------
    filename : str
        The filename to clean.
    """
    filename = filename.replace(" ", "_")
    filename = filename.replace(".", "_")
    filename = filename.replace(":", "-")
    filename = filename.replace("/", "-")
    filename = filename.replace("\\", "-")
    filename = filename.replace("?", "_")
    filename = filename.replace("*", "_")
    filename = filename.replace("<", "_smaller_")
    filename = filename.replace(">", "_greater_")
    filename = filename.replace("|", "-")
    filename = filename.replace('"', "_quote_")
    return filename



# Import the necessary PySide6 modules
import sys
from PySide6.QtWidgets import QApplication, QDialog, QProgressBar, QGridLayout, QLabel
from PySide6.QtCore import Qt, QCoreApplication, QThread, Signal
from PySide6.QtGui import QPixmap, QIcon
import os

sys.path.append(os.path.dirname(os.path.dirname(__file__)))
os.environ["QT_AUTO_SCREEN_SCALE_FACTOR"] = "1"

from importlib import resources, import_module
from nomad_camels import graphics

# Add nomad_camels/gui folder to the python path
sys.path.append(os.path.join(os.path.dirname(__file__), r"gui"))

# Import your main application form
# from main_window import MainWindow


# Create a new form for the loading screen
class LoadingScreen(QDialog):
    """
    A loading screen that appears when the application is starting up.
    """

    def __init__(self):
        super().__init__()
        self.setWindowTitle("NOMAD CAMELS - Loading...")
        self.setCursor(Qt.WaitCursor)
        layout = QGridLayout()
        self.setLayout(layout)
        self.setWindowFlags(Qt.CustomizeWindowHint | Qt.WindowTitleHint)
        self.setWindowIcon(QIcon(str(resources.files(graphics) / "camels_icon.png")))

        image = QPixmap()
        image.load(str(resources.files(graphics) / "CAMELS_vertical.png"))
        image = image.scaled(403, 308)
        image_label = QLabel()
        image_label.setPixmap(image)

        self.label = QLabel("loading...")

        self.progress_bar = QProgressBar(self)
        layout.addWidget(image_label, 0, 0)
        layout.addWidget(self.label, 1, 0)
        layout.addWidget(self.progress_bar, 2, 0)
        self.show()

    # A function to update the progress bar
    def set_progress(self, value):
        """
        Set the progress bar value
        """
        self.progress_bar.setValue(value)

    def set_text(self, text):
        """

        Parameters
        ----------
        text :


        Returns
        -------

        """
        self.label.setText(text)

    # Create a thread to import the packages


class ImportThread(QThread):
    """This thread imports the necessary packages for the main application to run."""

    update_progress = Signal(int)
    update_text = Signal(str)

    def __init__(self, package_list=None):
        super().__init__()
        self.package_list = package_list or []

    def run(self):
        """ """
        # Import your packages here
        package_list = self.package_list
        n = len(package_list) + 1
        for i, package in enumerate(package_list):
            self.update_progress.emit(int(i / n * 96))
            self.update_text.emit(f"loading {package}...")
            try:
                import_module(package)
            except ModuleNotFoundError:
                pass
            except AttributeError:
                pass
        self.update_text.emit("starting NOMAD CAMELS...")
        self.update_progress.emit(int((n - 1) / n * 96))
        from nomad_camels import MainApp_v2


# Show the loading screen and import your packages
def start_camels():
    import os

    app = QCoreApplication.instance()
    if app is None:
        # sys.argv += ['-platform', 'windows:darkmode=1']
        app = QApplication(sys.argv)

    # Create the loading screen
    loading_screen = LoadingScreen()
    loading_screen.show()
    import os.path

    file_dir = os.path.dirname(__file__)
    package_file = os.path.join(file_dir, "startup_packages.txt")
    if os.path.isfile(package_file):
        with open(package_file, "r", encoding="utf-8") as f:
            package_list = [x.rstrip() for x in f.readlines()]
    else:
        package_list = []

    # Start the thread and connect the progress signal
    thread = ImportThread(package_list)
    thread.update_progress.connect(loading_screen.set_progress)
    thread.update_text.connect(loading_screen.set_text)
    thread.start()
    while thread.isRunning():
        app.processEvents()
    with open(package_file, "w", encoding="utf-8") as f:
        packages = {key: val for key, val in sys.modules.items()}
        for i, (mod_name, mod) in enumerate(packages.items()):
            if mod_name.startswith("_") or mod is None:
                continue
            f.write(f"{mod_name}\n")
    from nomad_camels import MainApp_v2
    from nomad_camels.utility import exception_hook

    sys.excepthook = exception_hook.exception_hook
    main_window = MainApp_v2.MainWindow()
    loading_screen.hide()
    main_window.show()

    app.exec()


if __name__ == "__main__":
    start_camels()



""""""

import os
import pathlib

import databroker

from nomad_camels.utility import load_save_functions, update_camels
from nomad_camels.ui_widgets.warn_popup import WarnPopup


def make_yml(datapath, catalog_name="CAMELS_CATALOG", ask_restart=False):
    """
    Creates the yml file for the databroker (where it is looking for them) to
    configure a simple catalog for measurements with CAMELS.

    Parameters
    ----------
    datapath : str, path
        The path where the measurement data of the catalog should be saved to.
    catalog_name : str
         (Default value = 'CAMELS_CATALOG')
         The name, the catalog should have.
    """
    catalog_path = databroker.catalog_search_path()[0]
    if not isinstance(datapath, pathlib.Path):
        datapath = pathlib.Path(datapath)
    if not isinstance(catalog_path, pathlib.Path):
        catalog_path = pathlib.Path(catalog_path)
    if not os.path.isdir(catalog_path):
        os.makedirs(catalog_path)
    fname = (catalog_path / catalog_name).with_suffix(".yml")
    brokerpath = datapath / "databroker" / catalog_name
    if not os.path.isdir(brokerpath):
        os.makedirs(brokerpath)
    with open(fname, "w", encoding="utf-8") as f:
        f.write(
            "sources:\n"
            f"  {catalog_name}:\n"
            '    driver: "bluesky-msgpack-catalog"\n'
            "    args:\n"
            "      paths:\n"
            f'        - "{brokerpath.as_posix()}/*.msgpack"'
        )
    if catalog_name not in list(databroker.catalog):
        databroker.catalog.force_reload()
    if ask_restart and catalog_name not in list(databroker.catalog):
        WarnPopup(
            text="Could not load the databroker catalog, you might have to restart CAMELS",
            title="Databroker catalog not loaded",
        )
        update_camels.restart_camels(ask_restart=True)



from PySide6.QtWidgets import QTreeView, QAbstractItemView
from PySide6.QtCore import Signal, Qt
from PySide6.QtGui import QDropEvent, QKeyEvent, QPainter, QColor


class Drag_Drop_TreeView(QTreeView):
    """This Class is used for the protocol sequence. Most importantly it
    emits the dragdrop signal when something is dragged / dropped.

    Parameters
    ----------

    Returns
    -------

    """

    dragdrop = Signal()
    del_clicked = Signal()

    def __init__(self):
        super(Drag_Drop_TreeView, self).__init__()
        self.setHeaderHidden(True)
        self.setContextMenuPolicy(Qt.CustomContextMenu)
        self.setDragDropMode(QAbstractItemView.InternalMove)

    def dropEvent(self, e: QDropEvent) -> None:
        """

        Parameters
        ----------
        e: QDropEvent :


        Returns
        -------

        """
        super(Drag_Drop_TreeView, self).dropEvent(e)
        self.dragdrop.emit()

    def keyPressEvent(self, event: QKeyEvent) -> None:
        """

        Parameters
        ----------
        event: QKeyEvent :


        Returns
        -------

        """
        if event.key() == Qt.Key_Delete:
            self.del_clicked.emit()

    def paintEvent(self, event):
        super().paintEvent(event)
        if self.model() is not None and self.model().rowCount() == 0:
            painter = QPainter(self.viewport())
            painter.setPen(QColor(128, 128, 128))  # Gray color
            # Set font to be larger and bold
            font = self.font()
            font.setPointSize(font.pointSize() + 2)  # Make font larger
            font.setBold(True)
            painter.setFont(font)
            rect = self.viewport().rect()
            painter.drawText(rect, Qt.AlignCenter, "Right-click to add\nyour first step here")
 



from PySide6.QtWidgets import (
    QWidget,
    QGridLayout,
    QLabel,
    QComboBox,
    QLineEdit,
    QTextEdit,
)
from PySide6.QtCore import Qt, Signal
from nomad_camels.ui_widgets.add_remove_table import AddRemoveTable
from nomad_camels.main_classes.loop_step import Loop_Step, Loop_Step_Config
import json
import requests
from urllib.parse import urlparse
import re


class API_Call(Loop_Step):
    """
    A loop step that makes an API call.
    You can perform a CAMELS API call with more sophisticated control and better handling.
    Or you can perform an arbitrary API call.
    You always need to specify the host (ip address) and the port of the web server supplying the API.
    """

    def __init__(self, name="", parent_step=None, step_info=None, **kwargs):
        super().__init__(name, parent_step, step_info, **kwargs)
        self.step_type = "API Call"
        self.selected_camels_function_index = None
        if step_info is None:
            step_info = {}

        # Extract information from the step_info dictionary
        self.camels_function_parameters = step_info.get(
            "camels_function_parameters", None
        )
        # Get the host address
        self.host = step_info["host"] if "host" in step_info else ""
        # Get the port
        self.port = step_info["port"] if "port" in step_info else ""
        self.api_url = step_info["api_url"] if "api_url" in step_info else ""
        self.generic_api_method = (
            step_info["generic_api_method"]
            if "generic_api_method" in step_info
            else "GET"
        )
        self.api_type = step_info["api_type"] if "api_type" in step_info else "CAMELS"
        self.message_body = (
            step_info["message_body"] if "message_body" in step_info else ""
        )
        self.authentication_type = (
            step_info["authentication_type"]
            if "authentication_type" in step_info
            else ""
        )
        self.authentication_string = (
            step_info["authentication_string"]
            if "authentication_string" in step_info
            else ""
        )
        self.selected_camels_function_index = step_info.get(
            "selected_camels_function_index", 0
        )
        if self.selected_camels_function_index == "None":
            self.selected_camels_function_index = None
        self.camels_function_parameters = step_info.get(
            "camels_function_parameters", None
        )
        self.protocol_step_name = step_info.get("name", None)
        self.http_parameters = step_info.get("http_parameters", {})

    def get_protocol_string(self, n_tabs=1):
        """
        This function executes the API call and returns the results.
        """
        tabs = "\t" * n_tabs
        protocol_string = super().get_protocol_string(n_tabs)
        if self.api_type == "CAMELS":
            if not isinstance(self.message_body, dict):
                protocol_string += f"{tabs}message_body = ''\n"
            else:
                protocol_string += f"{tabs}message_body = helper_functions.evaluate_message_body({self.message_body}, eva, camels=True)\n"
            # Set port to empty string if it is not set

            protocol_string += (
                f"{tabs}api_result = helper_functions.execute_camels_api_call("
            )
            protocol_string += f"host='{self.host}', port='{self.port}', api_type='{self.api_type}', message_body=message_body, authentication_type='{self.authentication_type}', authentication_string='{self.authentication_string}', selected_camels_function_index={self.selected_camels_function_index}, camels_function_parameters={self.camels_function_parameters}"
            protocol_string += ")\n"
            protocol_string += f"{tabs}helper_functions.save_API_response_to_variable(api_result, namespace, protocol_step_name='{self.protocol_step_name}')\n"
        elif self.api_type == "Generic":
            if not isinstance(self.message_body, dict):
                protocol_string += f"{tabs}message_body = ''\n"
            else:
                protocol_string += f"{tabs}message_body = helper_functions.evaluate_message_body({self.message_body}, eva, camels=False)\n"
            protocol_string += (
                f"{tabs}api_result = helper_functions.execute_generic_api_call("
            )
            protocol_string += f"host='{self.host}', port='{self.port}', api_type='{self.api_type}', api_url='{self.api_url}', http_method='{self.generic_api_method}', message_body=message_body, authentication_type='{self.authentication_type}', authentication_string='{self.authentication_string}', http_parameters={self.http_parameters}"
            protocol_string += ")\n"
            protocol_string += f"{tabs}helper_functions.save_API_response_to_variable(api_result, namespace, protocol_step_name='{self.protocol_step_name}')\n"

        return protocol_string


class API_Call_Config(Loop_Step_Config):
    """
    A class to configure the API Call loop step.
    """

    def __init__(
        self,
        loop_step: API_Call,
        parent=None,
    ):
        super().__init__(parent, loop_step)
        self.loop_step = loop_step
        self.api_functions_list = {}

        # Label setup
        self.label_host = QLabel("Host:")
        self.label_port = QLabel("Port:")
        self.label_api_type = QLabel("API Type:")
        self.label_message_body = QLabel("Message Body:")
        self.label_authentication_type = QLabel("Authentication Type:")
        self.label_authentication_string = QLabel("Authentication:")
        self.label_camels_api_functions = QLabel("CAMELS API Functions:")
        self.label_api_url = QLabel("API URL:")
        self.label_generic_api_method = QLabel("HTTP Method:")
        self.label_table_parameters = QLabel("HTTP Parameters:")

        # Create combobox for API type
        self.combobox_camels_api_functions = None
        self.combobox_api_type = QComboBox()
        self.combobox_api_type.addItem("CAMELS")
        self.combobox_api_type.addItem("Generic")
        self.combobox_api_type.currentIndexChanged.connect(self.api_type_changed)

        # Create combobox for HTTP method
        self.combobox_generic_api_method = QComboBox()
        self.combobox_generic_api_method.addItem("GET")
        self.combobox_generic_api_method.addItem("POST")
        self.combobox_generic_api_method.addItem("DELETE")
        self.combobox_generic_api_method.addItem("PATCH")
        self.combobox_generic_api_method.addItem("PUT")
        self.combobox_generic_api_method.currentIndexChanged.connect(
            self.update_generic_api_method
        )

        # Line edit for API URL
        self.line_edit_api_url = QLineEdit()
        self.line_edit_api_url.setText(self.loop_step.api_url)
        self.line_edit_api_url.textChanged.connect(self.update_api_url)

        # Line edit host
        self.line_edit_host = QLineEdit()
        self.line_edit_host.setText(self.loop_step.host)
        self.line_edit_host.textChanged.connect(self.update_host)
        # Line edit port
        self.line_edit_port = QLineEdit()
        self.line_edit_port.setText(self.loop_step.port)
        self.line_edit_port.textChanged.connect(self.update_port)
        # Text edit post body
        self.text_edit_message_body = QTextEdit()
        self.text_edit_message_body.setText(self.loop_step.message_body)
        self.text_edit_message_body.textChanged.connect(self.update_message_body)

        # Variables table for the HTTP parameters
        self.table_http_parameters = AddRemoveTable(headerLabels=["Parameter", "Value"])
        self.table_http_parameters.setToolTip(
            "Add the parameters and their values for the HTTP request.\nThey are appended to the URL like\n<URL>?<parameter1>=<value1>&<parameter2>=<value2>&..."
        )
        self.table_http_parameters.table_model.itemChanged.connect(
            self.update_parameters_table
        )
        self.table_http_parameters.change_table_data(self.loop_step.http_parameters)
        self.table_http_parameters.table_model.rowsRemoved.connect(
            self.update_parameters_table
        )

        # Combobox for authentication type
        self.combobox_authentication_type = QComboBox()
        self.combobox_authentication_type.addItem("None")
        self.combobox_authentication_type.addItem("HTTP Basic")
        self.combobox_authentication_type.addItem("Bearer Token")
        self.combobox_authentication_type.setCurrentText(
            self.loop_step.authentication_type
        )
        self.combobox_authentication_type.currentIndexChanged.connect(
            self.authentication_type_changed
        )
        # Add tooltips to the combobox items
        self.combobox_authentication_type.setItemData(
            0, "No authentication required", Qt.ToolTipRole
        )
        self.combobox_authentication_type.setItemData(
            1, "Use format: username:password", Qt.ToolTipRole
        )
        self.combobox_authentication_type.setItemData(
            2, "Use format: Token", Qt.ToolTipRole
        )
        # Line edit authentication string
        self.line_edit_authentication_string = QLineEdit()
        self.line_edit_authentication_string.setEchoMode(QLineEdit.Password)
        self.line_edit_authentication_string.textChanged.connect(
            self.update_authentication_string
        )
        self.line_edit_authentication_string.setToolTip(
            "For HTTP Basic Authentication, use the format: username:password.\nFor Bearer Token Authentication, use the format: Token."
        )
        self.line_edit_authentication_string.setText(
            self.loop_step.authentication_string
        )

        # Combobox for CAMELS API functions
        self.CAMELS_functions_layout = QGridLayout()
        self.combobox_camels_api_functions = Combobox_CAMELS_functions()
        # Update the available CAMELS API functions when clicking on the combobox
        self.combobox_camels_api_functions.popupAboutToBeShown.connect(
            self.extract_all_api_functionality_from_json
        )
        self.combobox_camels_api_functions.currentIndexChanged.connect(
            self.update_camels_function_parameters
        )
        # Load the latest CAMELS function selection from the last time the protocol was closed
        if self.loop_step.selected_camels_function_index is not None:
            self.load_camels_function_on_bootup(
                self.loop_step.selected_camels_function_index
            )

        # Layout setup
        layout = self.layout()
        # Set host
        layout.addWidget(self.label_host, 1, 0)
        layout.addWidget(self.line_edit_host, 1, 1)
        # Set port
        layout.addWidget(self.label_port, 2, 0)
        layout.addWidget(self.line_edit_port, 2, 1)
        # Set API type
        layout.addWidget(self.label_api_type, 3, 0)
        layout.addWidget(self.combobox_api_type, 3, 1)
        # Set CAMELS API functions
        layout.addWidget(self.label_camels_api_functions, 4, 0)
        layout.addWidget(self.combobox_camels_api_functions, 4, 1)
        # Set Generic API URL
        layout.addWidget(self.label_api_url, 5, 0)
        layout.addWidget(self.line_edit_api_url, 5, 1)
        # Set genereic API HTTP method
        layout.addWidget(self.label_generic_api_method, 6, 0)
        layout.addWidget(self.combobox_generic_api_method, 6, 1)
        # Set CAMELS parameter layout
        camels_function_parameter_widget = QWidget()
        camels_function_parameter_widget.setLayout(self.CAMELS_functions_layout)
        layout.addWidget(camels_function_parameter_widget, 7, 0, 1, 2)

        # Set post body
        layout.addWidget(self.label_message_body, 12, 0)
        layout.addWidget(self.text_edit_message_body, 12, 1)

        # Set parameters table
        layout.addWidget(self.label_table_parameters, 15, 0)
        layout.addWidget(self.table_http_parameters, 16, 0, 1, 2)

        # Set authentication type
        layout.addWidget(self.label_authentication_type, 20, 0)
        layout.addWidget(self.combobox_authentication_type, 20, 1)
        self.combobox_authentication_type.setToolTip(
            "For HTTP Basic Authentication, use the format 'username:password'.\nFor Bearer Token Authentication, use the format: Token."
        )
        # Set authentication string
        layout.addWidget(self.label_authentication_string, 21, 0)
        layout.addWidget(self.line_edit_authentication_string, 21, 1)

        # Set the saved API type to the combobox
        self.combobox_api_type.setCurrentText(self.loop_step.api_type)
        self.combobox_api_type.currentIndexChanged.emit(
            self.combobox_api_type.currentIndex()
        )
        # Set the saved HTTP method to the combobox
        self.combobox_generic_api_method.setCurrentIndex(
            self.combobox_generic_api_method.findText(self.loop_step.generic_api_method)
        )

    def update_parameters_table(self):
        """
        Update the parameters table.
        """
        self.loop_step.http_parameters = self.table_http_parameters.update_table_data()

    def rows_removed_parameters_table(self):
        """
        Update the parameters table.
        """
        self.loop_step.http_parameters = self.table_http_parameters.update_table_data()

    def update_api_url(self):
        """
        Update the API URL.
        """
        self.loop_step.api_url = self.line_edit_api_url.text()
        # Check if the URL starts with a forward slash, if not add it
        if not self.loop_step.api_url.startswith("/"):
            self.loop_step.api_url = "/" + self.loop_step.api_url
        parsed_url = urlparse(self.loop_step.api_url)
        if parsed_url.scheme == "":
            api_url_with_schema = "http://" + self.loop_step.api_url
        parsed_url = urlparse(api_url_with_schema)
        if parsed_url.netloc == "":
            pass
        elif re.match(r"^[^:]+:\d+$", parsed_url.netloc):
            try:
                self.loop_step.host = parsed_url.netloc.split(":")[0]
            except ValueError:
                print(
                    "Failed to parse the host from the URL. Please check the URL format."
                )
            try:
                self.loop_step.port = parsed_url.netloc.split(":")[1]
            except IndexError:
                print(
                    "Failed to parse the port from the URL. Please check the URL format."
                )
            self.line_edit_host.setText(self.loop_step.host)
            self.line_edit_host.textChanged.emit(self.line_edit_host.text())
            self.line_edit_port.setText(self.loop_step.port)
            self.line_edit_port.textChanged.emit(self.line_edit_port.text())

    def update_host(self):
        """
        Update the host address.
        """
        self.loop_step.host = self.line_edit_host.text()

    def update_port(self):
        """
        Update the port.
        """
        self.loop_step.port = self.line_edit_port.text()

    def update_message_body(self):
        """
        Update the post body. Keeps the formatting."""
        self.loop_step.message_body = self.text_edit_message_body.toPlainText()

    def update_authentication_string(self):
        """
        Update the authentication string.
        """
        self.loop_step.authentication_string = (
            self.line_edit_authentication_string.text()
        )

    def update_generic_api_method(self):
        """ """
        self.loop_step.generic_api_method = (
            self.combobox_generic_api_method.currentText()
        )
        if self.loop_step.generic_api_method in {"GET", "DELETE"}:
            self.label_message_body.setVisible(False)
            self.text_edit_message_body.setVisible(False)
        elif self.loop_step.generic_api_method in {"POST", "PATCH", "PUT"}:
            self.label_message_body.setVisible(True)
            self.text_edit_message_body.setVisible(True)

    def api_type_changed(self):
        """ """
        self.loop_step.api_type = self.combobox_api_type.currentText()
        if self.loop_step.api_type == "CAMELS":
            self.label_camels_api_functions.setVisible(True)
            self.combobox_camels_api_functions.setVisible(True)
            self.label_api_url.setVisible(False)
            self.line_edit_api_url.setVisible(False)
            self.label_generic_api_method.setVisible(False)
            self.combobox_generic_api_method.setVisible(False)

            # Set authentication type and hide it
            self.combobox_authentication_type.setCurrentText("Bearer Token")
            self.label_authentication_type.setVisible(False)
            self.combobox_authentication_type.setVisible(False)

            if self.loop_step.host and self.loop_step.port:
                self.extract_all_api_functionality_from_json()
            if self.loop_step.selected_camels_function_index is not None:
                self.load_camels_function_on_bootup(
                    self.loop_step.selected_camels_function_index
                )

        elif self.loop_step.api_type == "Generic":
            self.label_authentication_type.setVisible(True)
            self.combobox_authentication_type.setVisible(True)
            self.label_api_url.setVisible(True)
            self.line_edit_api_url.setVisible(True)
            self.label_generic_api_method.setVisible(True)
            self.combobox_generic_api_method.setVisible(True)
            # emit signal of the combobox to update the method
            self.combobox_generic_api_method.setCurrentIndex(
                self.combobox_generic_api_method.findText(
                    self.loop_step.generic_api_method
                )
            )
            self.combobox_generic_api_method.currentIndexChanged.emit(
                self.combobox_generic_api_method.currentIndex()
            )
            self.combobox_camels_api_functions.setVisible(False)
            self.label_camels_api_functions.setVisible(False)
            if self.combobox_camels_api_functions is not None:
                # Temporarily disconnect the signal
                self.combobox_camels_api_functions.currentIndexChanged.disconnect(
                    self.update_camels_function_parameters
                )
                # Clear the combobox
                self.combobox_camels_api_functions.clear()
                # Reconnect the signal
                self.combobox_camels_api_functions.currentIndexChanged.connect(
                    self.update_camels_function_parameters
                )

                # Hide camels function parameters
                for i in reversed(range(self.CAMELS_functions_layout.count())):
                    self.CAMELS_functions_layout.itemAt(i).widget().setParent(None)

    def authentication_type_changed(self):
        """ """
        self.loop_step.authentication_type = (
            self.combobox_authentication_type.currentText()
        )

    def extract_all_api_functionality_from_json(self):
        """ """
        # Make the API call to get the OpenAPI schema
        host = self.loop_step.host
        port = self.loop_step.port
        if host == "":
            self.combobox_camels_api_functions.clear()
            self.combobox_camels_api_functions.addItem(
                "Please enter a host address first."
            )
            return
        if port == "":
            self.combobox_camels_api_functions.clear()
            self.combobox_camels_api_functions.addItem("Please enter a port first.")
            return
        try:
            json_response = requests.get(f"http://{host}:{port}/openapi.json")
        except requests.exceptions.ConnectionError as e:
            print(
                "Could not connect to the server.\nMake sure the server is running and that the host and port are correct."
            )
            self.combobox_camels_api_functions.clear()
            self.combobox_camels_api_functions.addItem(
                "Could not connect to the server."
            )
            return
        # Parse the JSON response
        openapi_schema = json.loads(json_response.text)
        paths = openapi_schema.get("paths", {})
        # Extract the function names and descriptions
        api_functions = []
        for path, methods in paths.items():
            for method, details in methods.items():
                summary = details.get("summary", "No summary available")
                operation_id = details.get("operationId", "Unnamed function")
                description = details.get("description", "No description available")
                parameters = details.get("parameters", [])
                api_functions.append(
                    {
                        "summary": summary,
                        "method": method.upper(),
                        "path": path,
                        "operation_id": operation_id,
                        "description": description,
                        "parameters": parameters,
                    }
                )
        self.api_functions_list = api_functions
        # Update the combobox with the API functions
        # Temporarily disconnect the signal
        self.combobox_camels_api_functions.currentIndexChanged.disconnect(
            self.update_camels_function_parameters
        )
        # Clear the combobox
        self.combobox_camels_api_functions.clear()
        for func in api_functions:
            self.combobox_camels_api_functions.addItem(func["summary"])
            index = self.combobox_camels_api_functions.count() - 1
            self.combobox_camels_api_functions.setItemData(
                index, func["description"], Qt.ToolTipRole
            )
        # Reconnect the signal
        self.combobox_camels_api_functions.currentIndexChanged.connect(
            self.update_camels_function_parameters
        )
        if self.loop_step.selected_camels_function_index is not None:
            self.combobox_camels_api_functions.setCurrentIndex(
                int(self.loop_step.selected_camels_function_index)
            )
            try:
                selected_function = self.api_functions_list[
                    self.loop_step.selected_camels_function_index
                ]
            except IndexError:
                self.loop_step.selected_camels_function_index = -1
                print(
                    "Could not load the CAMELS function. Check your connection to the server."
                )
                return
            if selected_function["parameters"]:
                # clear the layout
                for i in reversed(range(self.CAMELS_functions_layout.count())):
                    self.CAMELS_functions_layout.itemAt(i).widget().setParent(None)
                self.CAMELS_functions_layout.addWidget(QLabel("Parameters:"), 0, 0)
                self.CAMELS_functions_layout.addWidget(QLabel("Value:"), 0, 1)
                self.camels_function_parameters = {}
                for i, parameter in enumerate(selected_function["parameters"]):
                    self.CAMELS_functions_layout.addWidget(
                        QLabel(parameter["name"]), i + 1, 0
                    )
                    line_edit_parameter = QLineEdit()
                    self.CAMELS_functions_layout.addWidget(
                        line_edit_parameter, i + 1, 1
                    )
                    # match the parameter name with the value
                    if (
                        self.loop_step.camels_function_parameters is not None
                        and self.loop_step.camels_function_parameters != "None"
                    ):
                        for (
                            parameter_name,
                            value,
                        ) in self.loop_step.camels_function_parameters.items():
                            if parameter_name == parameter["name"]:
                                line_edit_parameter.setText(value)
                    line_edit_parameter.textChanged.connect(
                        lambda text, param_name=parameter[
                            "name"
                        ]: self.update_parameter_value(param_name, text)
                    )

    def update_camels_function_parameters(self, index=None):
        """ """
        selected_index = self.combobox_camels_api_functions.currentIndex()
        if index is not None:
            selected_index = index
        if len(self.api_functions_list) == 0:
            return
        selected_function = self.api_functions_list[selected_index]
        if selected_function["method"] == "GET":
            self.label_message_body.setVisible(False)
            self.text_edit_message_body.setVisible(False)
        elif selected_function["method"] == "POST":
            self.label_message_body.setVisible(True)
            self.text_edit_message_body.setVisible(True)
        if selected_function["parameters"]:
            # clear the layout
            for i in reversed(range(self.CAMELS_functions_layout.count())):
                self.CAMELS_functions_layout.itemAt(i).widget().setParent(None)
            self.CAMELS_functions_layout.addWidget(QLabel("Parameters:"), 0, 0)
            self.CAMELS_functions_layout.addWidget(QLabel("Value:"), 0, 1)
            self.camels_function_parameters = {}
            for i, parameter in enumerate(selected_function["parameters"]):
                self.CAMELS_functions_layout.addWidget(
                    QLabel(parameter["name"]), i + 1, 0
                )
                line_edit_parameter = QLineEdit()
                self.CAMELS_functions_layout.addWidget(line_edit_parameter, i + 1, 1)
                line_edit_parameter.textChanged.connect(
                    lambda text, param_name=parameter[
                        "name"
                    ]: self.update_parameter_value(param_name, text)
                )
        else:
            # clear the layout
            for i in reversed(range(self.CAMELS_functions_layout.count())):
                self.CAMELS_functions_layout.itemAt(i).widget().setParent(None)
        self.loop_step.selected_camels_function_index = (
            self.combobox_camels_api_functions.currentIndex()
        )

    def load_camels_function_on_bootup(self, index):
        """ """
        self.extract_all_api_functionality_from_json()
        self.combobox_camels_api_functions.setCurrentIndex(index)
        try:
            self.update_camels_function_parameters(index)
        except KeyError as e:
            print(
                "Could not load the CAMELS function. Check your connection to the server."
            )
        if (
            self.loop_step.camels_function_parameters is not None
            and self.loop_step.camels_function_parameters != "None"
        ):
            for (
                parameter_name,
                value,
            ) in self.loop_step.camels_function_parameters.items():
                for i in range(self.CAMELS_functions_layout.count()):
                    if (
                        self.CAMELS_functions_layout.itemAt(i).widget().text()
                        == parameter_name
                    ):
                        self.CAMELS_functions_layout.itemAt(i + 1).widget().setText(
                            value
                        )

    def update_parameter_value(self, parameter_name, value):
        """ """
        self.camels_function_parameters[parameter_name] = value
        self.loop_step.camels_function_parameters = self.camels_function_parameters


class Combobox_CAMELS_functions(QComboBox):
    popupAboutToBeShown = Signal()

    def showPopup(self):
        self.popupAboutToBeShown.emit()
        super().showPopup()



import os.path
from PySide6.QtWidgets import (
    QApplication,
    QDialog,
    QLabel,
    QLineEdit,
    QGridLayout,
    QDialogButtonBox,
    QComboBox,
    QTextEdit,
)
from PySide6.QtCore import Qt

import yaml

from nomad_camels.ui_widgets.path_button_edit import Path_Button_Edit
from nomad_camels.nomad_integration import nomad_communication


class EntrySelector(QDialog):
    def __init__(self, parent=None, selection="Sample"):
        super().__init__(parent)
        self.reloading = False
        self.setWindowTitle(f"Select {selection}")
        self.entry_metadata = []
        self.entry_names = []
        self.entry_uploads = []
        self.entry_types = []
        self.entry_data = []

        label_entry_scope = QLabel("Entry Scope:")
        self.comboBox_entry_scope = QComboBox()
        self.comboBox_entry_scope.addItems(["shared", "user"])
        self.comboBox_entry_scope.setCurrentText("user")
        self.comboBox_entry_scope.setToolTip(
            "shared: Entries that belong to you or are shared with you.\nuser: Entries that belong to you."
        )

        self.button_box = QDialogButtonBox(
            QDialogButtonBox.Ok | QDialogButtonBox.Cancel
        )
        self.button_box.accepted.connect(self.accept)
        self.button_box.rejected.connect(self.reject)

        label_upload = QLabel("Upload:")
        self.upload_box = QComboBox()

        label_entry_type = QLabel("Entry Type:")
        self.entry_type_box = QComboBox()

        label_entry = QLabel("Entry:")
        self.entry_box = QComboBox()

        self.entry_info = QTextEdit()
        self.entry_info.setTextInteractionFlags(
            Qt.TextSelectableByKeyboard | Qt.TextSelectableByMouse
        )

        layout = QGridLayout()
        layout.addWidget(label_entry_scope, 0, 0)
        layout.addWidget(self.comboBox_entry_scope, 0, 1)
        layout.addWidget(label_upload, 1, 0)
        layout.addWidget(self.upload_box, 1, 1)
        layout.addWidget(label_entry_type, 2, 0)
        layout.addWidget(self.entry_type_box, 2, 1)
        layout.addWidget(label_entry, 10, 0)
        layout.addWidget(self.entry_box, 10, 1)
        layout.addWidget(self.entry_info, 0, 2, 12, 1)
        layout.addWidget(self.button_box, 20, 0, 1, 3)
        self.setLayout(layout)

        self.update_entries()
        self.entry_filtering()
        self.entry_change()
        self.upload_box.currentTextChanged.connect(self.entry_filtering)
        self.entry_type_box.currentTextChanged.connect(self.entry_filtering)
        self.entry_box.currentTextChanged.connect(self.entry_change)
        self.comboBox_entry_scope.currentTextChanged.connect(self.update_entries)

        self.return_data = {}

        self.adjustSize()

    def update_entries(self):
        self.setCursor(Qt.WaitCursor)
        self.setEnabled(False)
        entries = nomad_communication.get_entries(
            self.parent, owner=self.comboBox_entry_scope.currentText()
        )["data"]
        if not entries:
            raise Exception("No Entries found!")
        self.reloading = True
        self.entry_metadata.clear()
        self.entry_names.clear()
        self.entry_uploads.clear()
        self.entry_types.clear()
        self.entry_data.clear()
        for entry in entries:
            if "archive" not in entry:
                continue
            arch = entry["archive"]
            if "data" not in arch:
                continue
            self.entry_data.append(arch["data"])
            self.entry_metadata.append(arch["metadata"])
            self.entry_names.append(arch["metadata"]["entry_name"])
            self.entry_types.append(arch["metadata"]["entry_type"])
            if "upload_name" in arch["metadata"]:
                self.entry_uploads.append(arch["metadata"]["upload_name"])
            else:
                self.entry_uploads.append(arch["metadata"]["upload_id"])

        self.upload_box.clear()
        self.upload_box.addItems(sorted(list(set(self.entry_uploads))))

        self.entry_type_box.clear()
        self.entry_type_box.addItems(sorted(list(set(self.entry_types))))

        self.entry_box.clear()
        self.entry_box.addItems(sorted(self.entry_names))
        self.entry_filtering()
        self.entry_change()
        self.reloading = False
        self.setEnabled(True)
        self.setCursor(Qt.ArrowCursor)

    def entry_filtering(self):
        if self.reloading:
            return
        upload = self.upload_box.currentText()
        entry_type = self.entry_type_box.currentText()
        entries = []
        for i, entry in enumerate(self.entry_names):
            if upload == self.entry_uploads[i] and entry_type == self.entry_types[i]:
                entries.append(entry)
        self.entry_box.clear()
        self.entry_box.addItems(entries)

    def entry_change(self):
        if self.reloading:
            return
        self.entry_info.setText(yaml.dump(self.get_current_entry_data()))

    def get_current_entry_data(self, include_metadata=False):
        entry = self.entry_box.currentText()
        for i, ent in enumerate(self.entry_names):
            if ent == entry:
                if include_metadata:
                    data = dict(self.entry_data[i])
                    data["NOMAD_entry_metadata"] = dict(self.entry_metadata[i])
                    return data
                return dict(self.entry_data[i])
        return {}

    def accept(self):
        from nomad_camels.nomad_integration.nomad_communication import nomad_url

        self.return_data = self.get_current_entry_data(True)
        if "name" not in self.return_data and "Name" not in self.return_data:
            self.return_data["name"] = self.entry_box.currentText().split(".")[0]
        self.return_data["ELN-service"] = "nomad"
        self.return_data["full_identifier"] = (
            f'{nomad_url}/upload/id/{self.return_data["NOMAD_entry_metadata"]["upload_id"]}/entry/id/{self.return_data["NOMAD_entry_metadata"]["entry_id"]}'
        )
        lab_id = self.return_data.get("lab_id", None)
        if lab_id:
            self.return_data["identifier"] = self.return_data["lab_id"]
        else:
            self.return_data["identifier"] = (
                f'{nomad_url}/upload/id/{self.return_data["NOMAD_entry_metadata"]["upload_id"]}/entry/id/{self.return_data["NOMAD_entry_metadata"]["entry_id"]}'
            )
        super().accept()



"""Helping functions for loading/saving. These are not in `load_save_functions`,
so they may also be imported by modules that are imported by `load_save_functions`."""

from nomad_camels.frontpanels.plot_definer import Plot_Info, Fit_Info


def load_plots(plots: list, plot_data):
    """
    Loads the data from `plot_data` into `Plot_Info` objects. Then adds them to
    the `plots` list.

    Parameters
    ----------
    plots : list
        the list, where to add the plot information
    plot_data : list[dict]
        list of dictionaries containing the necessary information for Plot_Info

    Returns
    -------
    plots : list
        the modified list with the loaded plots added
    """
    plots.clear()
    for plot_dict in plot_data:
        plot = Plot_Info()
        plots.append(plot)
        for k, v in plot_dict.items():
            if k == "all_fit":
                fit = Fit_Info()
                plot.__dict__[k] = fit
                for k2, v2 in v.items():
                    fit.__dict__[k2] = v2
            elif k == "fits":
                fits = []
                plot.__dict__[k] = fits
                for fit_dict in v:
                    fit = Fit_Info()
                    fits.append(fit)
                    for k2, v2 in fit_dict.items():
                        fit.__dict__[k2] = v2
            else:
                plot.__dict__[k] = v
    return plots



from PySide6.QtWidgets import QTableView, QWidget, QVBoxLayout, QPushButton
from PySide6.QtGui import QStandardItemModel, QStandardItem, QPainter, QColor, QIcon
from PySide6.QtCore import Qt, Signal
from nomad_camels.utility import variables_handling

from importlib import resources
from nomad_camels import graphics


class VariableTable(QTableView):
    """ """

    def __init__(self, parent=None, protocol=None, editable_names=True, variables=None):
        super().__init__(parent)
        self.model = QStandardItemModel()
        self.setModel(self.model)
        self.model.setHorizontalHeaderLabels(["Name", "Value", "Data-Type"])
        self.model.itemChanged.connect(self.check_variable)
        self.editable_names = editable_names
        self.protocol = protocol
        self.variables = variables
        if protocol:
            self.set_protocol(protocol)
        elif variables:
            for var in sorted(variables):
                self.append_variable(var, str(variables[var]), unique=False)

    def paintEvent(self, event):
        super().paintEvent(event)
        if self.model.rowCount() == 0:
            painter = QPainter(self.viewport())
            painter.setPen(QColor(128, 128, 128))  # Gray color

            # Set font to be larger and bold
            font = self.font()
            font.setPointSize(font.pointSize() + 2)  # Make font larger
            font.setBold(True)
            painter.setFont(font)

            rect = self.viewport().rect()
            painter.drawText(rect, Qt.AlignCenter, "Define your variables here.")

    def set_protocol(self, protocol):
        """ """
        self.protocol = protocol
        for var in sorted(self.protocol.variables):
            self.append_variable(var, str(self.protocol.variables[var]), unique=False)

    def append_variable(self, name="name", value="value", unique=True):
        """ """
        if unique:
            name = self.get_unique_name(name)
        name_item = QStandardItem(name)
        value_item = QStandardItem(value)
        type_item = QStandardItem(variables_handling.check_data_type(value))
        name_item.setEditable(self.editable_names)
        type_item.setEditable(False)
        self.model.appendRow([name_item, value_item, type_item])

    def check_variable(self):
        """ """
        ind = self.selectedIndexes()
        if ind:
            ind = ind[0]
        else:
            return
        item = self.model.itemFromIndex(ind)
        if ind.column() == 0:
            variables_handling.check_variable_name(item.text(), parent=self)
        if ind.column() == 0 and item.text() in self.protocol.variables:
            new_name = self.get_unique_name(item.text())
            item.setText(new_name)
            raise Exception("Variable names must be unique!")
        if ind.column() == 1:
            d_type = variables_handling.check_data_type(item.text())
            self.model.item(ind.row(), 2).setText(d_type)
        self.update_variables()

    def update_variables(self):
        """ """
        variables = {}
        for i in range(self.model.rowCount()):
            name = self.model.item(i, 0).text()
            value = variables_handling.get_data(self.model.item(i, 1).text())
            variables.update({name: value})
        if self.editable_names:
            self.protocol.variables = variables
            variables_handling.protocol_variables = self.protocol.variables
        else:
            return variables

    def get_unique_name(self, name="name"):
        """ """
        i = 1
        while name in self.protocol.variables:
            if "_" not in name:
                name += f"_{i}"
            else:
                name = f'{name.split("_")[0]}_{i}'
            i += 1
        return name

    def clear(self):
        """ """
        self.model.clear()
        self.model.setHorizontalHeaderLabels(["Name", "Value", "Data-Type"])
        self.model.itemChanged.connect(self.check_variable)
        self.update_variables()


class VariableBox(QWidget):
    new_values_signal = Signal(dict)
    closing = Signal()

    def __init__(
        self, parent=None, protocol=None, editable_names=True, variables=None, name=""
    ):
        super().__init__(parent)
        name = name or (protocol.name if protocol else "protocol")
        self.setWindowTitle(f"Live variable control - {name} - NOMAD CAMELS")
        self.setWindowIcon(QIcon(str(resources.files(graphics) / "camels_icon.png")))
        self.layout = QVBoxLayout()
        self.setLayout(self.layout)
        self.table = VariableTable(
            protocol=protocol,
            editable_names=editable_names,
            parent=self,
            variables=variables,
        )
        self.layout.addWidget(self.table)
        self.button = QPushButton("update values")
        self.button.clicked.connect(self.update_values)
        self.layout.addWidget(self.button)
        # Disable the close button
        self.setWindowFlags(self.windowFlags() & ~Qt.WindowCloseButtonHint)

    def update_values(self):
        variables = self.table.update_variables()
        self.new_values_signal.emit(variables)
        return variables

    def closeEvent(self, event):
        # Emit the closing signal
        self.closing.emit()
        # Accept the event to close the window
        event.accept()
